# cython: embedsignature=True

import operator

cimport cython
from cpython.tuple cimport (PyTuple_GET_ITEM, PyTuple_GetItem, PyTuple_GetSlice,
                            PyTuple_GET_SIZE, PyTuple_New, PyTuple_SetItem)
from cpython.list cimport PyList_Append, PyList_AsTuple
from cpython.ref cimport Py_INCREF
from cpython.object cimport PyObject
# from cpython.bytearray cimport PyByteArray_FromStringAndSize

# to support older versions of Cython which do not include this header
cdef extern from "Python.h":
    bytearray PyByteArray_FromStringAndSize(char *string, Py_ssize_t len)


from libc.string cimport memcpy
from libc.math cimport fabs
from libc.stdlib cimport malloc, free

from ms_deisotope._c.averagine cimport mass_charge_ratio
from ms_peak_picker._c.peak_set cimport PeakBase

cimport numpy as np

np.import_array()


@cython.cdivision
cdef double ppm_error(double x, double y):
    return (x - y) / y


cdef:
    object neutral_mass_getter
    object mz_getter

cdef double epsilon = 1e-3

neutral_mass_getter = operator.attrgetter("neutral_mass")
mz_getter = operator.attrgetter('mz')


cpdef _Index_reconstructor(object neutral_mass, object mz=None):
    if isinstance(neutral_mass, int):
        return _Index._create(neutral_mass, mz)
    elif isinstance(neutral_mass, bytearray):
        return _Index_from_bytes(neutral_mass)
    else:
        raise TypeError(type(neutral_mass))


cpdef _Index_to_bytes(_Index index):
    cdef:
        size_t j, m
        char* buff
        np.uint64_t n
        bytearray out

    m = sizeof(np.uint64_t) * 2
    buff = <char*>malloc(sizeof(char) * m)
    j = 0
    n = index.neutral_mass
    memcpy(&buff[j], &n, sizeof(np.uint64_t))
    j += sizeof(np.uint64_t)
    n = index.mz
    memcpy(&buff[j], &n, sizeof(np.uint64_t))
    j += sizeof(np.uint64_t)
    out = PyByteArray_FromStringAndSize(<char*>buff, m)
    free(buff)
    return out


cpdef _Index_from_bytes(bytearray view):
    cdef:
        size_t j
        np.uint64_t val
        char* buff
        _Index inst
    inst = _Index._create(0, 0)
    j = 0
    buff = view
    memcpy(&val, &buff[j], sizeof(np.uint64_t))
    j += sizeof(np.uint64_t)
    inst.neutral_mass = val
    memcpy(&val, &buff[j], sizeof(np.uint64_t))
    j += sizeof(np.uint64_t)
    inst.mz = val
    return inst

@cython.freelist(100000)
cdef class _Index(object):
    """
    Stores the ordered index of an object under sort by `mz` and
    sort by `neutral_mass`

    Attributes
    ----------
    mz : int
        Index of this object (or its parent), ordered by mz
    neutral_mass : int
        Index of this object (or its parent), ordered by neutral mass
    """
    def __init__(self, neutral_mass=0, mz=0):
        self.neutral_mass = neutral_mass
        self.mz = mz

    def __repr__(self):
        return "%d|%d" % (self.neutral_mass, self.mz)

    def clone(self):
        return _Index._create(self.neutral_mass, self.mz)

    def __reduce__(self):
        return _Index_reconstructor, (self.neutral_mass, self.mz)

    @staticmethod
    cdef _Index _create(size_t neutral_mass, size_t mz):
        cdef:
            _Index inst
        inst = _Index.__new__(_Index)
        inst.neutral_mass = neutral_mass
        inst.mz = mz
        return inst


@cython.freelist(1000000)
cdef class EnvelopePair(object):

    def __init__(self, mz, intensity):
        self.mz = mz
        self.intensity = intensity

    def __getitem__(self, i):
        if i == 0:
            return self.mz
        elif i == 1:
            return self.intensity
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.mz
        yield self.intensity

    def __reduce__(self):
        return EnvelopePair, (self.mz, self.intensity,)

    cpdef bint _eq(self, EnvelopePair other):
        return (abs(self.mz - other.mz) < epsilon) and (
            abs(self.intensity - other.intensity) < epsilon)

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))


    @staticmethod
    cdef EnvelopePair _create(double mz, double intensity):
        cdef:
            EnvelopePair inst
        inst = EnvelopePair.__new__(EnvelopePair)
        inst.mz = mz
        inst.intensity = intensity
        return inst

    def __repr__(self):
        return "(%0.4f, %0.2f)" % (self.mz, self.intensity)


@cython.boundscheck(False)
cpdef bytearray Envelope_to_bytes(Envelope self):
    cdef:
        size_t i, j, m
        EnvelopePair pair
        char* buff
        np.uint32_t n
        np.float64_t val
        bytearray out

    n = self.get_size()
    m = sizeof(np.uint32_t) + sizeof(np.float64_t) * n * 2
    buff = <char*>malloc(sizeof(char) * m)
    j = 0
    memcpy(&buff[j], &n, sizeof(np.uint32_t))
    j += sizeof(np.uint32_t)
    for i in range(n):
        pair = self.getitem(i)
        val = pair.mz
        memcpy(&buff[j], &val, sizeof(np.float64_t))
        j += sizeof(np.float64_t)
        val = pair.intensity
        memcpy(&buff[j], &val, sizeof(np.float64_t))
        j += sizeof(np.float64_t)
    out = PyByteArray_FromStringAndSize(<char*>buff, m)
    free(buff)
    return out


cpdef Envelope_from_bytes(bytearray view):
    cdef:
        size_t i, j, m
        EnvelopePair pair
        np.uint32_t n
        np.float64_t val
        char* buff
        tuple result
    j = 0
    buff = view
    memcpy(&n, &buff[j], sizeof(np.uint32_t))
    j += sizeof(np.uint32_t)
    result = PyTuple_New(n)
    for i in range(n):
        pair = EnvelopePair._create(0, 0)
        memcpy(&pair.mz, &buff[j], sizeof(np.float64_t))
        j += sizeof(np.float64_t)
        memcpy(&pair.intensity, &buff[j], sizeof(np.float64_t))
        j += sizeof(np.float64_t)
        Py_INCREF(pair)
        PyTuple_SetItem(result, i, pair)
    return Envelope._create(result)

@cython.boundscheck(False)
cpdef Envelope_reconstructor(object value):
    if isinstance(value, tuple):
        return Envelope._create(value)
    elif isinstance(value, bytearray):
        return Envelope_from_bytes(value)
    else:
        raise TypeError(type(value))


@cython.freelist(100000)
cdef class Envelope(object):
    """
    Represents a sequence of (mz, intensity) pairs which store peak positions
    that were matched. Since these peaks may later be mutated by subtraction,
    this structure stores copies of the numbers

    Supports the Sequence protocol, and is read-only.

    Attributes
    ----------
    pairs : tuple of tuple of (float, float)
        list of (mz, intensity) pairs
    """
    def __init__(self, pairs):
        self.pairs = tuple(map(lambda x: EnvelopePair(*x), pairs))

    def __getitem__(self, i):
        return self.pairs[i]

    def __iter__(self):
        return iter(self.pairs)

    def __repr__(self):
        return "[%s]" % (', '.join("(%0.4f, %0.2f)" % tuple(t) for t in self))

    def __len__(self):
        return len(self.pairs)

    cdef inline size_t get_size(self):
        return PyTuple_GET_SIZE(self.pairs)

    cdef inline EnvelopePair getitem(self, size_t i):
        return <EnvelopePair>PyTuple_GetItem(self.pairs, i)

    cpdef bint _eq(self, Envelope other):
        cdef:
            size_t i, n
            EnvelopePair a, b

        n = self.get_size()
        if n != other.get_size():
            return False

        i = 0
        for i in range(n):
            a = self.getitem(i)
            b = other.getitem(i)
            if a != b:
                return False
        return True

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))

    cpdef Envelope clone(self):
        cdef:
            tuple duplicate
            size_t i, n
            EnvelopePair pair, dup_pair
        n = self.get_size()
        duplicate = PyTuple_New(n)
        for i in range(n):
            pair = self.getitem(i)
            dup_pair = EnvelopePair._create(pair.mz, pair.intensity)
            Py_INCREF(dup_pair)
            PyTuple_SetItem(duplicate, i, dup_pair)
        return Envelope._create(duplicate)

    def __reduce__(self):
        return (
            Envelope_reconstructor,
            (Envelope_to_bytes(self),))
            # (self.pairs, ))

    @staticmethod
    cdef Envelope _create(tuple pairs):
        cdef:
            Envelope inst
        inst = Envelope.__new__(Envelope)
        inst.pairs = pairs
        return inst


cdef class DeconvolutedPeak(PeakBase):
    """
    Represent a single deconvoluted peak which represents an aggregated isotopic
    pattern collapsed to its monoisotopic peak, with a known charge state

    Attributes
    ----------
    a_to_a2_ratio : float
        Ratio of intensities of A peak to A+2 peak
    average_mass : float
        The averaged neutral mass of the composition, the weighted average
        of the envelope peaks.
    charge : int
        The signed charge state of the isotopic pattern
    envelope : Envelope
        The sequence of (mz, intensity) pairs which map to this peak
    full_width_at_half_max : float
        The averaged full width at half max of this isotopic pattern
    index : _Index
        The position of this peak in different orderings
    intensity : float
        The summed height of the peaks in the isotopic pattern this peak captures
    most_abundant_mass : float
        The neutral mass of the most abundant peak in the isotopic pattern this peak captures
    mz : float
        The mass-charge-ratio of the monoisotopic peak
    neutral_mass : float
        The neutral mass of the monoisotopic peak
    score : float
        An assigned value describing the quality of this peak's fit. The semantics of this score
        depends upon the scoring function
    signal_to_noise : float
        The average signal-to-noise ratio of the peaks in the isotopic pattern this peak captures
    """
    def __init__(self, neutral_mass, intensity, charge, signal_to_noise, index, full_width_at_half_max,
                 a_to_a2_ratio=0, most_abundant_mass=0, average_mass=0, score=0,
                 envelope=(), mz=0, fit=None, chosen_for_msms=False, area=0):
        if index is None:
            index = _Index._create(0, 0)
        elif index == -1:
            index = _Index._create(0, 0)
        if envelope is None:
            envelope = ()
        self.neutral_mass = neutral_mass
        self.intensity = intensity
        self.signal_to_noise = signal_to_noise
        self.index = index
        self.full_width_at_half_max = full_width_at_half_max
        self.charge = charge
        self.a_to_a2_ratio = a_to_a2_ratio
        self.most_abundant_mass = most_abundant_mass
        self.average_mass = average_mass
        self.score = score
        if not isinstance(envelope, Envelope):
            self.envelope = Envelope(envelope)
        else:
            self.envelope = envelope
        self.mz = mz or mass_charge_ratio(self.neutral_mass, self.charge)
        self.fit = fit
        self.chosen_for_msms = chosen_for_msms
        self.area = area

    property index:
        def __get__(self):
            return self._index

        def __set__(self, value):
            if isinstance(value, _Index):
                self._index = value
            elif value >= 0:
                self._index = _Index._create(value, value)
            else:
                self._index = _Index._create(0, 0)

    def __hash__(self):
        return hash((self.mz, self.intensity, self.charge))

    cpdef PeakBase clone(self):
        return DeconvolutedPeak(self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
                                self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
                                self.most_abundant_mass, self.average_mass, self.score,
                                self.envelope, self.mz, self.fit, self.chosen_for_msms, self.area)

    def __reduce__(self):
        return DeconvolutedPeak, (self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
                                  self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
                                  self.most_abundant_mass, self.average_mass, self.score,
                                  self.envelope, self.mz, self.fit, self.chosen_for_msms, self.area)

    cpdef bint _eq(self, DeconvolutedPeak other):
        return (abs(self.neutral_mass - other.neutral_mass) < epsilon) and (
            abs(self.intensity - other.intensity) < epsilon) and (
            self.charge == other.charge) and (
            abs(self.score - other.score) < epsilon)

    def __richcmp__(self, other, int code):
        if code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))

    def __repr__(self):
        return ("DeconvolutedPeak(a_to_a2_ratio={self.a_to_a2_ratio}, area={self.area}, "
            "average_mass={self.average_mass}, charge={self.charge}, chosen_for_msms={self.chosen_for_msms}, "
            "envelope={self.envelope}, full_width_at_half_max={self.full_width_at_half_max}, index={self.index}, "
            "intensity={self.intensity}, most_abundant_mass={self.most_abundant_mass}, mz={self.mz}, "
            "neutral_mass={self.neutral_mass}, score={self.score}, signal_to_noise={self.signal_to_noise})").format(self=self)


    @staticmethod
    cdef DeconvolutedPeak _create_simple(double neutral_mass, double intensity, int charge,
                                         double score, double mz, Envelope envelope):
        cdef:
            DeconvolutedPeak inst
        inst = DeconvolutedPeak.__new__(DeconvolutedPeak)
        inst.neutral_mass = neutral_mass
        inst.intensity = intensity
        inst.charge = charge
        inst.score = score
        inst.signal_to_noise = score
        inst.mz = mz
        inst.envelope = envelope
        inst.index = _Index._create(0, 0)
        inst.full_width_at_half_max = 0
        inst.a_to_a2_ratio = 0
        inst.most_abundant_mass = 0
        inst.average_mass = 0
        inst.area = 0

        return inst


cdef class DeconvolutedPeakSolution(DeconvolutedPeak):
    """
    Extends :class:`DeconvolutedPeak` to also include a reference to
    the :class:`IsotopicFitRecord` instance it is derived from, and optionally an
    object which the fit derives from.

    Attributes
    ----------
    fit : IsotpicFitRecord
        The isotopic pattern fit used to construct this peak
    solution : object
        Representation of the "source" of the isotopic fit
    """
    def __init__(self, solution, fit, *args, **kwargs):
        self.solution = solution
        self.fit = fit
        super(DeconvolutedPeakSolution, self).__init__(*args, **kwargs)

    cpdef PeakBase clone(self):
        return DeconvolutedPeakSolution(
            self.solution, self.fit, self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.chosen_for_msms, self.area)

    def __reduce__(self):
        return DeconvolutedPeakSolution, (
            self.solution, self.fit, self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.chosen_for_msms, self.area)

    def __iter__(self):
        yield self.solution
        yield self
        yield self.fit


cdef class DeconvolutedPeakSet:
    """
    Represents a collection of :class:`DeconvolutedPeak` instances under multiple orderings.

    Attributes
    ----------
    peaks : tuple of DeconvolutedPeak
        Collection of peaks ordered by `neutral_mass`
    _mz_ordered: tuple of DeconvolutedPeak
        Collection of peaks ordered by `mz`
    """
    def __init__(self, peaks):
        self.peaks = tuple(peaks)
        self._mz_ordered = None
        self.indexed = False

    cpdef reindex(self):
        """
        Updates the :attr:`index` of each peak in `self` and updates the
        sorted order.

        Returns
        -------
        self: DeconvolutedPeakSet
        """
        cdef:
            size_t i, n
            DeconvolutedPeak peak
        self.peaks = tuple(sorted(self.peaks, key=neutral_mass_getter))
        self._mz_ordered = tuple(sorted(self.peaks, key=mz_getter))
        n = PyTuple_GET_SIZE(self.peaks)
        i = 0
        for i in range(n):
            peak = self.getitem(i)
            peak._index = _Index._create(0, 0)
            peak._index.neutral_mass = i
        i = 0
        for i in range(n):
            peak = <DeconvolutedPeak>PyTuple_GET_ITEM(self._mz_ordered, i)
            peak._index.mz = i
        self.indexed = True
        return self

    def __iter__(self):
        return iter(self.peaks)

    def __len__(self):
        return self.get_size()

    cdef size_t get_size(self):
        return PyTuple_GET_SIZE(self.peaks)

    def __repr__(self):
        return "<DeconvolutedPeakSet %d Peaks>" % (len(self))

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__class__(tuple(p.clone() for p in self.peaks[item]))
        return self.peaks[item]

    def __eq__(self, other):
        try:
            return self.peaks == other.peaks
        except AttributeError:
            return tuple(self) == tuple(other)

    def __ne__(self, other):
        try:
            return self.peaks != other.peaks
        except AttributeError:
            return tuple(self) != tuple(other)

    def __iter__(self):
        return iter(self.peaks)

    cpdef DeconvolutedPeakSet clone(self):
        return self.copy()

    cpdef DeconvolutedPeakSet copy(self):
        cdef:
            tuple duplicate_peaks
            DeconvolutedPeak peak
            size_t i, n
        n = self.get_size()
        duplicate_peaks = PyTuple_New(n)
        for i in range(n):
            peak = self.getitem(i)
            peak = peak.clone()
            Py_INCREF(peak)
            PyTuple_SetItem(duplicate_peaks, i, peak)
        return self.__class__(duplicate_peaks)

    def __reduce__(self):
        return self.__class__, (self.peaks, )

    cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=1e-5, bint use_mz=False):
        '''Find the peak that best matches ``neutral_mass`` within ``error_tolerance`` mass accuracy ppm.

        If ``use_mz`` is True, instead of matching neutral masses, match peaks using m/z instead.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        error_tolerance: double
            The PPM error tolerance to apply
        use_mz: bool
            Whether to search using m/z instead of neutral mass

        Returns
        -------
        DeconvolutedPeak
            The found peak, or None if no peak is found
        '''
        if use_mz:
            return binary_search_mz(self._mz_ordered, neutral_mass, error_tolerance)
        else:
            return binary_search_neutral_mass(self.peaks, neutral_mass, error_tolerance)

    cpdef DeconvolutedPeak has_peak(self, double neutral_mass, double error_tolerance=1e-5, bint use_mz=False):
        '''Find the peak that best matches ``neutral_mass`` within ``error_tolerance`` mass accuracy ppm.

        If ``use_mz`` is True, instead of matching neutral masses, match peaks using m/z instead.

        If :attr:`indexed` is not :const:`True`, then :meth:`reindex` will be called.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        error_tolerance: double
            The PPM error tolerance to apply
        use_mz: bool
            Whether to search using m/z instead of neutral mass

        Returns
        -------
        DeconvolutedPeak
            The found peak, or None if no peak is found
        '''
        if not self.indexed:
            self.reindex()
        return self._has_peak(neutral_mass, error_tolerance, use_mz)

    cdef DeconvolutedPeak getitem(self, size_t i):
        return <DeconvolutedPeak>PyTuple_GET_ITEM(self.peaks, i)

    cdef tuple getslice(self, size_t start, size_t end):
        return <tuple>PyTuple_GetSlice(self.peaks, start, end)

    cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=1e-5):
        '''Find all peaks that match ``neutral_mass`` within ``error_tolerance`` mass accuracy ppm.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        error_tolerance: double
            The PPM error tolerance to apply

        Returns
        -------
        tuple of DeconvolutedPeak
            The found peaks
        '''
        cdef:
            double lo, hi, lo_err, hi_err
            int lo_ix, hi_ix
            DeconvolutedPeak lo_peak
            DeconvolutedPeak hi_peak
        lo = neutral_mass - neutral_mass * tolerance
        hi = neutral_mass + neutral_mass * tolerance
        lo_peak = binary_search_nearest_neutral_mass(self.peaks, lo, &lo_err)
        hi_peak = binary_search_nearest_neutral_mass(self.peaks, hi, &hi_err)
        lo_ix = lo_peak._index.neutral_mass
        if lo_ix < PyTuple_GET_SIZE(self.peaks) and abs(
                ppm_error(lo_peak.neutral_mass, neutral_mass)) > tolerance:
            lo_ix += 1
        hi_ix = hi_peak._index.neutral_mass + 1
        if hi_ix != 0 and abs(
                ppm_error(hi_peak.neutral_mass, neutral_mass)) > tolerance:
            hi_ix -= 1
        return <tuple>PyTuple_GetSlice(self.peaks, lo_ix, hi_ix)

    cdef DeconvolutedPeak _get_nearest_peak(self, double neutral_mass, double* errout):
        '''Find the peak nearest to ``neutral_mass``, regardless of error.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        errout: double*
            A pointer to store the mass error in

        Returns
        -------
        DeconvolutedPeak
            The nearest peak
        '''
        cdef DeconvolutedPeak peak
        peak = binary_search_nearest_neutral_mass(self.peaks, neutral_mass, errout)
        return peak

    def get_nearest_peak(self, double neutral_mass):
        '''Find the peak nearest to ``neutral_mass``, regardless of error.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for

        Returns
        -------
        DeconvolutedPeak
            The nearest peak
        double
            The error between ``neutral_mass`` and the found peak
        '''
        cdef:
            DeconvolutedPeak peak
            double errout
        peak = self._get_nearest_peak(neutral_mass, &errout)
        return peak, errout

    def between(self, m1, m2, tolerance=1e-5, use_mz=False):
        """Retrieve a :class:`DeconvolutedPeakSet` containing all the peaks
        whose mass is between ``m1`` and ``m2``.

        If ``use_mz`` is :const:`True` then search by m/z instead of mass

        Parameters
        ----------
        m1 : float
            The lower mass limit
        m2 : float
            The upper mass limit
        use_mz: bool
            Whether to search for m/z instead of neutral mass

        Returns
        -------
        DeconvolutedPeakSet
        """
        acc = []
        collecting = False
        value = None
        if not use_mz:
            getter = operator.attrgetter("neutral_mass")
            iterator = self.peaks
        else:
            getter = operator.attrgetter("mz")
            iterator = self._mz_ordered
        for peak in iterator:
            value = getter(peak)
            if not collecting and value >= m1 and value < m2:
                collecting = True
            elif collecting and value > m2:
                break

            if collecting:
                acc.append(peak.clone())

        return self.__class__(acc).reindex()

cdef double INF
INF = float('inf')


cdef DeconvolutedPeak _sweep_solution_neutral_mass(tuple array, double value, size_t mid, double tolerance):
    cdef:
        size_t best_index, i, n
        double best_error, abs_error
        DeconvolutedPeak target

    best_index = mid
    best_error = INF
    n = PyTuple_GET_SIZE(array)

    i = 0
    while mid - i >= 0 and i <= mid:
        target = <DeconvolutedPeak>array[mid - i]
        abs_error = abs(ppm_error(value, target.neutral_mass))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid - i
                best_error = abs_error
        else:
            break
        i += 1
    i = 1
    while (mid + i) < (n - 1):
        target = <DeconvolutedPeak>array[mid + i]
        abs_error = abs(ppm_error(value, target.neutral_mass))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid + i
                best_error = abs_error
        else:
            break
        i += 1
    if best_error == INF:
        return None
    else:
        return array[best_index]


@cython.cdivision(True)
cdef double _ppm_error(double x, double y) nogil:
    return (x - y) / y


cdef int _binary_search(double* array, double target, double error_tolerance, size_t n, size_t* out) nogil:
    cdef:
        size_t lo, hi, mid, i, j
        size_t best_index
        double err, found_mass
        double best_error, abs_error

    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        found_mass = array[mid]
        if fabs(_ppm_error(found_mass, target)) < error_tolerance:
            best_index = mid
            best_error = INF
            i = 0
            while mid - i >= 0 and i <= mid:
                found_mass = array[mid - i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid - i
                        best_error = abs_error
                else:
                    break
                i += 1
            i = 1
            while (mid + i) < (n - 1):
                found_mass = array[mid + i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid + i
                        best_error = abs_error
                else:
                    break
                i += 1
            out[0] = best_index
            if best_error == INF:
                return 3
            return 0
        elif hi - lo == 1:
            out[0] = 0
            return 1
        elif found_mass > target:
            hi = mid
        else:
            lo = mid
    out[0] = 0
    return 2


cdef int _binary_search_interval(double* array, double target, double error_tolerance, size_t n, size_t* start, size_t* end) nogil:
    cdef:
        size_t lo, hi, mid, i, j
        size_t best_index
        double err, found_mass
        double best_error, abs_error

    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        found_mass = array[mid]
        if fabs(_ppm_error(found_mass, target)) < error_tolerance:
            best_index = mid
            best_error = INF
            i = 0
            while mid - i >= 0 and i <= mid:
                found_mass = array[mid - i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid - i
                        best_error = abs_error
                else:
                    break
                i += 1
            start[0] = mid - i + 1
            i = 1
            while (mid + i) < (n - 1):
                found_mass = array[mid + i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid + i
                        best_error = abs_error
                else:
                    break
                i += 1
            end[0] = mid + i
            return 0
        elif hi - lo == 1:
            start[0] = 0
            end[0] = 0
            return 1
        elif found_mass > target:
            hi = mid
        else:
            lo = mid
    start[0] = 0
    end[0] = 0
    return 2


cdef DeconvolutedPeak binary_search_neutral_mass(tuple peak_set, double neutral_mass, double error_tolerance):
    cdef:
        size_t lo, hi, mid, i, j
        double err, found_mass
        DeconvolutedPeak found_peak

    lo = 0
    hi = PyTuple_GET_SIZE(peak_set)

    while hi != lo:
        mid = (hi + lo) / 2
        found_peak = <DeconvolutedPeak>peak_set[mid]
        found_mass = found_peak.neutral_mass

        if abs(ppm_error(found_mass, neutral_mass)) < error_tolerance:
            found_peak = _sweep_solution_neutral_mass(peak_set, neutral_mass, mid, error_tolerance)
            return found_peak
        elif hi - lo == 1:
            return None
        elif found_mass > neutral_mass:
            hi = mid
        else:
            lo = mid


cdef DeconvolutedPeak _sweep_solution_mz(tuple array, double value, size_t mid, double tolerance):
    cdef:
        size_t best_index, i, n
        double best_error, abs_error
        DeconvolutedPeak target

    best_index = mid
    best_error = INF
    n = PyTuple_GET_SIZE(array)

    i = 0
    while mid - i >= 0 and i <= mid:
        target = <DeconvolutedPeak>array[mid - i]
        abs_error = abs(ppm_error(value, target.mz))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid - i
                best_error = abs_error
        else:
            break
        i += 1
    i = 1
    while (mid + i) < (n - 1):
        target = <DeconvolutedPeak>array[mid + i]
        abs_error = abs(ppm_error(value, target.mz))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid + i
                best_error = abs_error
        else:
            break
        i += 1
    if best_error == INF:
        return None
    else:
        return array[best_index]


cdef DeconvolutedPeak binary_search_mz(tuple peak_set, double mz, double error_tolerance):
    cdef:
        size_t lo, hi, mid, i, j
        double err, found_mass
        DeconvolutedPeak found_peak

    lo = 0
    hi = len(peak_set)

    while hi != lo:
        mid = (hi + lo) / 2
        found_peak = <DeconvolutedPeak>peak_set[mid]
        found_mass = found_peak.mz

        if abs(ppm_error(found_mass, mz)) < error_tolerance:
            found_peak = _sweep_solution_mz(peak_set, mz, mid, error_tolerance)
            return found_peak
        elif hi - lo == 1:
            return None
        elif found_mass > mz:
            hi = mid
        else:
            lo = mid


cdef DeconvolutedPeak _sweep_nearest_match_neutral_mass(tuple array, double value, size_t lo, size_t hi, double* errout):
    cdef:
        size_t i
        size_t best_index
        double best_error, err
        double v

    best_error = float('inf')
    best_index = -1
    for i in range(hi - lo):
        i += lo
        v = array[i].neutral_mass
        err = abs(v - value)
        if err < best_error:
            best_error = err
            best_index = i
    errout[0] = best_error
    return array[best_index]


cdef DeconvolutedPeak binary_search_nearest_neutral_mass(tuple peak_set, double neutral_mass, double* errout):
    cdef:
        size_t lo, hi, mid, i, j
        double err, found_mass
        DeconvolutedPeak found_peak

    lo = 0
    hi = len(peak_set)
    while hi != lo:
        mid = (hi + lo) / 2
        found_peak = <DeconvolutedPeak>peak_set[mid]
        found_mass = found_peak.neutral_mass

        if abs(ppm_error(found_mass, neutral_mass)) < 1:
            found_peak = _sweep_nearest_match_neutral_mass(peak_set, neutral_mass, lo, hi, errout)
            return found_peak
        elif hi - lo == 1:
            errout[0] = found_mass - neutral_mass
            return found_peak
        elif found_mass > neutral_mass:
            hi = mid
        else:
            lo = mid


def convert(self):
    return DeconvolutedPeak(
        self.neutral_mass, self.intensity, self.charge,
        self.signal_to_noise, -1, self.full_width_at_half_max,
        self.a_to_a2_ratio, self.most_abundant_mass, self.average_mass,
        self.score, Envelope(self.envelope), self.mz, None, self.chosen_for_msms,
        self.area)


@cython.boundscheck(False)
@cython.wraparound(False)
def decode_envelopes(np.ndarray[cython.floating, ndim=1, mode='c'] array):
    cdef:
        list envelope_list, current_envelope
        tuple converted
        Envelope envelope_obj
        size_t i, n
        cython.floating a, b
    envelope_list = []
    current_envelope = []

    i = 0
    n = array.shape[0]
    while i < n:
        a = array[i]
        b = array[i + 1]
        i += 2
        if a == 0 and b == 0:
            if current_envelope is not None:
                if current_envelope:
                    converted = tuple(current_envelope)
                    envelope_obj = Envelope._create(converted)
                    envelope_list.append(envelope_obj)
                current_envelope = []
        else:
            current_envelope.append(EnvelopePair._create(a, b))
    converted = tuple(current_envelope)
    envelope_obj = Envelope._create(converted)
    envelope_list.append(envelope_obj)
    return envelope_list


INTERVAL_INDEX_SIZE = 0

cdef class DeconvolutedPeakSetIndexed(DeconvolutedPeakSet):
    def __init__(self, peaks):
        self.neutral_mass_array = NULL
        self.mz_array = NULL
        self.interval_index = NULL
        super(DeconvolutedPeakSetIndexed, self).__init__(peaks)

    def __dealloc__(self):
        self._release_buffers()

    def _release_buffers(self):
        if self.neutral_mass_array != NULL:
            free(self.neutral_mass_array)
            self.neutral_mass_array = NULL
        if self.mz_array != NULL:
            free(self.mz_array)
            self.mz_array = NULL
        if self.interval_index != NULL:
            free_index_list(self.interval_index)
            self.interval_index = NULL

    def set_interval_index_size(self, index_size):
        global INTERVAL_INDEX_SIZE
        INTERVAL_INDEX_SIZE = index_size
        self.reindex()

    cdef void _build_index_arrays(self):
        cdef:
            size_t i, n
            DeconvolutedPeak peak
        n = PyTuple_GET_SIZE(self.peaks)
        self._size = n
        self._release_buffers()
        self.neutral_mass_array = <double*>malloc(sizeof(double) * n)
        self.mz_array = <double*>malloc(sizeof(double) * n)

        for i in range(n):
            peak = self.getitem(i)
            self.neutral_mass_array[i] = peak.neutral_mass
            self.mz_array[peak._index.mz] = peak.mz

        # The interpolating interval index is just a bit slower than
        # the normal binary search, likely due to cache friendliness.
        if n > 2 and INTERVAL_INDEX_SIZE > 0:
            if self.interval_index != NULL:
                free_index_list(self.interval_index)
                self.interval_index = NULL
            interval_index = <index_list*>malloc(sizeof(index_list))
            build_interval_index(self, interval_index, INTERVAL_INDEX_SIZE)
            if check_index(interval_index) != 0:
                free_index_list(interval_index)
            else:
                self.interval_index = interval_index

    cpdef reindex(self):
        super(DeconvolutedPeakSetIndexed, self).reindex()
        self._build_index_arrays()
        return self

    cdef DeconvolutedPeak _has_peak(self, double neutral_mass, double error_tolerance=1e-5, bint use_mz=False):
        '''Find the peak that best matches ``neutral_mass`` within ``error_tolerance`` mass accuracy ppm.

        If ``use_mz`` is True, instead of matching neutral masses, match peaks using m/z instead.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        error_tolerance: double
            The PPM error tolerance to apply
        use_mz: bool
            Whether to search using m/z instead of neutral mass

        Returns
        -------
        DeconvolutedPeak
            The found peak, or None if no peak is found
        '''
        cdef:
            int status
            size_t i, n, s
            DeconvolutedPeak peak

        n = self._size
        if use_mz:
            status = _binary_search(self.mz_array, neutral_mass, error_tolerance, n, &i)
            if status != 0:
                return None
            else:
                return <DeconvolutedPeak>PyTuple_GET_ITEM(self._mz_ordered, i)
        if n == 0:
            return None
        if self.interval_index != NULL:
            find_search_interval(self.interval_index, neutral_mass, &s, &n)
            status = _binary_search_with_hint(self.neutral_mass_array, neutral_mass, error_tolerance, n, s, &i)
        else:
            status = _binary_search(self.neutral_mass_array, neutral_mass, error_tolerance, n, &i)
        if status != 0:
            return None
        peak = self.getitem(i)
        if abs((peak.neutral_mass - neutral_mass) / neutral_mass) < error_tolerance:
            return peak
        else:
            return None

    def _test_interval(self, double neutral_mass, double tolerance=1e-5):
        cdef:
            int status
            size_t n, start, end
        n = self._size
        status = _binary_search_interval(
            self.neutral_mass_array, neutral_mass, tolerance, n, &start, &end)
        return status, start, end

    cpdef tuple all_peaks_for(self, double neutral_mass, double tolerance=1e-5):
        '''Find all peaks that match ``neutral_mass`` within ``error_tolerance`` mass accuracy ppm.

        Parameters
        ----------
        neutral_mass: double
            The mass to search for
        error_tolerance: double
            The PPM error tolerance to apply

        Returns
        -------
        tuple of DeconvolutedPeak
            The found peaks
        '''
        cdef:
            int status
            size_t n, s, start, end
        n = self._size
        if not self.indexed:
            self.reindex()
        if n == 0:
            return ()
        if self.interval_index != NULL:
            find_search_interval(self.interval_index, neutral_mass, &s, &n)
            status = _binary_search_interval_with_hint(
                self.neutral_mass_array, neutral_mass, tolerance, n, s, &start, &end)
        else:
            status = _binary_search_interval(
                self.neutral_mass_array, neutral_mass, tolerance, n, &start, &end)
        if status != 0:
            return ()
        return <tuple>PyTuple_GetSlice(self.peaks, start, end)

    cdef int _interval_for(self, double neutral_mass, double tolerance, size_t* start, size_t* end) nogil:
        cdef:
            int status
            size_t n
        n = self._size
        status = _binary_search_interval(
            self.neutral_mass_array, neutral_mass, tolerance, n, start, end)
        return status

    def test_interval(self, double value):
        cdef:
            int status
            size_t start, end
            size_t i

        status = find_search_interval(self.interval_index, value, &start, &end)
        return start, end, interpolate_index(self.interval_index, value)

    def find_interval_for(self, double value):
        return interpolate_index(self.interval_index, value)

    def check_interval(self, size_t i):
        cdef:
            index_cell cell
        cell = self.interval_index.index[i]
        return cell



cdef int _binary_search_with_hint(double* array, double target, double error_tolerance, size_t n, size_t hint, size_t* out) nogil:
    '''Performs a best-matching binary search with a hint at the lower bound as well as the upper bound.

    The best-matching step occurs once a match is found, iterating forwards and backwards from the initial
    match until an index is found that minimizes the parts-per-million error with ``target``

    The binary search starts with the lower bound ``lo = hint`` instead of ``lo = 0``. The upper bound
    ``hi = n`` can be used as a hint to instruct the search to only consider a small segment of the array.

    Parameters
    ----------
    array: double*
        The array of doubles to search within
    target: double
        The value to search for
    error_tolerance: double
        The parts-per-million error tolerance to use to select solutions
    n: size_t
        The upper limit into ``array`` to search, usually the size of the array
        but may also be made lower to reduce the space to search
    hint: size_t
        The lower limit into ``array`` to search. Leaving this as 0
        will make this behave the same as :func:`_binary_search`.
    out: size_t*
        The best matching index
    '''
    cdef:
        size_t lo, hi, mid, i, j
        size_t best_index
        double err, found_mass
        double best_error, abs_error

    lo = hint
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        found_mass = array[mid]
        if fabs(_ppm_error(found_mass, target)) < error_tolerance:
            best_index = mid
            best_error = INF
            i = 0
            while mid - i >= 0 and i <= mid:
                found_mass = array[mid - i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid - i
                        best_error = abs_error
                else:
                    break
                i += 1
            i = 1
            while (mid + i) < (n - 1):
                found_mass = array[mid + i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid + i
                        best_error = abs_error
                else:
                    break
                i += 1
            out[0] = best_index
            if best_error == INF:
                return 3
            return 0
        elif hi - lo == 1:
            out[0] = 0
            return 1
        elif found_mass > target:
            hi = mid
        else:
            lo = mid
    out[0] = 0
    return 2


cdef int _binary_search_interval_with_hint(double* array, double target, double error_tolerance, size_t n, size_t hint, size_t* start, size_t* end) nogil:
    cdef:
        size_t lo, hi, mid, i, j
        size_t best_index
        double err, found_mass
        double best_error, abs_error

    lo = hint
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        found_mass = array[mid]
        if fabs(_ppm_error(found_mass, target)) < error_tolerance:
            best_index = mid
            best_error = INF
            i = 0
            while mid - i >= 0 and i <= mid:
                found_mass = array[mid - i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid - i
                        best_error = abs_error
                else:
                    break
                i += 1
            start[0] = mid - i + 1
            i = 1
            while (mid + i) < (n - 1):
                found_mass = array[mid + i]
                abs_error = fabs(_ppm_error(target, found_mass))
                if abs_error < error_tolerance:
                    if abs_error < best_error:
                        best_index = mid + i
                        best_error = abs_error
                else:
                    break
                i += 1
            end[0] = mid + i
            return 0
        elif hi - lo == 1:
            start[0] = 0
            end[0] = 0
            return 1
        elif found_mass > target:
            hi = mid
        else:
            lo = mid
    start[0] = 0
    end[0] = 0
    return 2


cdef int build_interval_index(DeconvolutedPeakSet peaks, index_list* index, size_t index_size):
    cdef:
        double* linear_spacing
        double current_value, err, next_value
        size_t i, start_i, end_i, index_i, peaks_size
        DeconvolutedPeak peak
    peaks_size = peaks.get_size()
    if peaks_size > 0:
        index.low = peaks.getitem(0).neutral_mass
        index.high = peaks.getitem(peaks_size - 1).neutral_mass
    else:
        index.low = 0
        index.high = 1
    index.size = index_size
    linear_spacing = build_linear_spaced_array(
        index.low,
        index.high,
        index_size)

    index.index = <index_cell*>malloc(sizeof(index_cell) * index_size)

    for index_i in range(index_size):
        current_value = linear_spacing[index_i]
        peak = peaks._get_nearest_peak(current_value, &err)
        if peaks_size > 0:
            start_i = peak._index.neutral_mass
            if index_i > 0:
                start_i = index.index[index_i - 1].end - 1
            if index_i == index_size - 1:
                end_i = peaks_size - 1
            else:
                next_value = linear_spacing[index_i + 1]
                i = peak._index.neutral_mass
                while i < peaks_size:
                    peak = peaks.getitem(i)
                    if abs(current_value - peak.neutral_mass) > abs(next_value - peak.neutral_mass):
                        break
                    i += 1
                end_i = i
        else:
            start_i = 0
            end_i = 0
        index.index[index_i].center_value = current_value
        index.index[index_i].start = start_i
        index.index[index_i].end = end_i

    free(linear_spacing)
    return 0


cdef double* build_linear_spaced_array(double low, double high, size_t n):
    cdef:
        double* array
        double delta
        size_t i

    delta = (high - low) / (n - 1)

    array = <double*>malloc(sizeof(double) * n)

    for i in range(n):
        array[i] = low + i * delta
    return array


cdef void free_index_list(index_list* index):
    free(index.index)
    free(index)


cdef int check_index(index_list* index) nogil:
    if index.size == 0:
        return 1
    elif (index.high - index.low) == 0:
        return 2
    else:
        return 0


cdef size_t interpolate_index(index_list* index, double value):
    cdef:
        double v
        size_t i
    v = (((value - index.low) / (index.high - index.low)) * index.size)
    i = <size_t>v
    return i


cdef int find_search_interval(index_list* index, double value, size_t* start, size_t* end):
    cdef:
        size_t i
    if value > index.high:
        i = index.size - 1
    elif value < index.low:
        i = 0
    else:
        i = interpolate_index(index, value)
    if i > 0:
        if i < index.size:
            start[0] = index.index[i - 1].start
        else:
            # if we're at index.size or greater, act as if we're at the last
            # cell of the index
            start[0] = index.index[index.size - 2].start
    else:
        # if somehow the index were negative, this could end badly.
        start[0] = index.index[i].start
    if i >= (index.size - 1):
        end[0] = index.index[index.size - 1].end + 1
    else:
        end[0] = index.index[i + 1].end + 1
    return 0



'''MGF is a simple human-readable format for MS/MS data. It
allows storing MS/MS peak lists and exprimental parameters.

This module provides :class:`MGFLoader`, a :class:`~.RandomAccessScanSource`
implementation.

The parser is based on :mod:`pyteomics.mgf`.
'''

from pyteomics import mgf
import numpy as np

from six import string_types as basestring

from .scan import (
    ScanFileMetadataBase, RandomAccessScanSource, ScanDataSource,
    PrecursorInformation, _FakeGroupedScanIteratorImpl,
    ChargeNotProvided)

from .metadata.file_information import (
    FileInformation, MS_MSn_Spectrum)


class _MGFParser(mgf.IndexedMGF):

    def parse_charge(self, charge_text, list_only=False):
        '''Pyteomics _parse_charge is very general-purpose, and
        can't be sped up, so we specialize it here.'''
        try:
            if not list_only:
                return int(charge_text.replace('+', ''))
            return list(map(self.parse_charge, charge_text.split(" ")))
        except Exception:
            if '-' in charge_text:
                return int(charge_text.replace("-", '')) * -1
            raise


class _MGFMetadata(ScanFileMetadataBase):
    """Objects implementing this interface can describe the original source
    files, instrument configuration, and data processing parameters used to
    create the current spectral data file.

    Patterned after the provenance features of mzML that could also be mapped
    onto mzXML and other complete vendor readers.
    """

    def file_description(self):
        '''Describe the file and its components, as well
        as any content types it has.

        Returns
        -------
        :class:`~.FileInformation`
        '''
        finfo = FileInformation()
        finfo.add_content("centroid spectrum")
        finfo.add_content(MS_MSn_Spectrum)
        if isinstance(self.source_file, basestring):
            finfo.add_file(self.source_file)
        elif hasattr(self.source_file, 'name'):
            finfo.add_file(self.source_file.name)
        return finfo

    def instrument_configuration(self):
        '''Describe the different instrument components and configurations used
        to acquire scans in this run.

        Returns
        -------
        :class:`list` of :class:`~.InstrumentInformation`
        '''
        return super(_MGFMetadata, self).data_processing()

    def data_processing(self):
        '''Describe any preprocessing steps applied to the data described by this
        instance.

        Returns
        -------
        :class:`list` of :class:`~.DataProcessingInformation`
        '''
        return super(_MGFMetadata, self).data_processing()


class MGFInterface(ScanDataSource):
    '''Provides a basic set of widely used MASCOT Generic File (MGF)
    data accessor mechanisms. Because MGF files lack any form of standardization,
    no strong guarantees of correctness can be made.

    This dialect does not know how to use the charge column of the peak data
    section, see :class:`~.ProcessedMGFLoader`.
    '''

    def _scan_arrays(self, scan):
        """Returns raw data arrays for m/z and intensity

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        mz: np.array
            An array of m/z values for this scan
        intensity: np.array
            An array of intensity values for this scan
        """
        try:
            return scan['m/z array'], scan["intensity array"]
        except KeyError:
            return np.array([]), np.array([])

    def _ms_level(self, scan):
        return 2

    def _scan_title(self, scan):
        """Returns a verbose name for this scan, if one
        were stored in the file. Usually includes both the
        scan's id string, as well as information about the
        original file and format.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return scan['params']["title"].strip('.')

    def _scan_id(self, scan):
        """Returns the scan's id string, a unique
        identifier for this scan in the context of
        the data file it is recordered in

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        str
        """
        return scan['params']["title"].strip('.')

    def _scan_time(self, scan):
        try:
            return float(scan['params']['rtinseconds']) / 60.0
        except KeyError:
            return -1

    def _is_profile(self, scan):
        return False

    def _precursor_information(self, scan):
        mz, intensity = scan['params']['pepmass']
        charge = scan['params'].get('charge', [ChargeNotProvided])[0]
        pinfo = PrecursorInformation(
            mz, intensity, charge, source=self,
            product_scan_id=self._scan_id(scan),
            defaulted=True, orphan=True)
        return pinfo

    def _polarity(self, scan):
        pinfo = self._precursor_information(scan)
        if pinfo is not None:
            if pinfo.charge:
                if pinfo.charge > 0:
                    return 1
                return -1
            return 1
        return 1

    def _activation(self, scan):
        return None

    def _scan_index(self, scan):
        """Returns the base 0 offset from the start
        of the data file in number of scans to reach
        this scan.

        If the original format does not natively include
        an index value, this value may be computed from
        the byte offset index.

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        int
        """
        try:
            return self._title_to_index[self._scan_title(scan)]
        except KeyError:
            try:
                return self._title_to_index[self._scan_title(scan) + '.']
            except KeyError:
                return -1
        return -1

    def _annotations(self, scan):
        annots = dict()
        params = scan['params']
        for key, value in params.items():
            if key in ("pepmass", "charge", "title", "rtinseconds"):
                continue
            else:
                try:
                    value = float(value)
                except ValueError:
                    if value == 'None':
                        value = None
            annots[key] = value
        return annots


class MGFLoader(MGFInterface, RandomAccessScanSource, _MGFMetadata):
    """Reads scans from MASCOT Generic File (MGF) Format files. Provides both iterative
    and random access.

    .. note::
        If the file is not sorted by retention time, :meth:`get_scan_by_time` and any
        other time-based accessors will fail.

    Attributes
    ----------
    source_file: str
        Path to file to read from.
    source: pyteomics.mgf.MGFBase
        Underlying scan data source
    header: dict
        Any top-of-the-file parameters
    """
    def __init__(self, source_file, encoding='utf-8', use_index=True, **kwargs):
        self.source_file = source_file
        self.encoding = encoding
        self._use_index = use_index
        self._source = self._create_parser()
        self.initialize_scan_cache()
        self.make_iterator()
        self._title_to_index = self._prepare_index_lookup()

    def _prepare_index_lookup(self):
        title_to_index = dict()
        for i, key in enumerate(self.index):
            title_to_index[key] = i
        return title_to_index

    @property
    def header(self):
        '''Any top-of-the-file parameters

        Returns
        -------
        dict
        '''
        return self._source.header

    def __reduce__(self):
        return self.__class__, (self.source_file, self.encoding, self._use_index, )

    def has_msn_scans(self):
        return True

    def has_ms1_scans(self):
        return False

    def _create_parser(self):
        if self._use_index:
            return _MGFParser(self.source_file, read_charges=False,
                              convert_arrays=1, encoding=self.encoding)
        return mgf.MGF(self.source_file, read_charges=False,
                       convert_arrays=1, encoding=self.encoding)

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        try:
            return self.scan_cache[scan_id]
        except KeyError:
            pass
        try:
            scan = self.source.get_spectrum(scan_id)
        except KeyError:
            scan = self.source.get_spectrum(scan_id + '.')
        scan = self._make_scan(scan)
        self.scan_cache[scan_id] = scan
        return scan

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        if not self._use_index:
            raise TypeError("This method requires the index. Please pass `use_index=True` during initialization")
        id_str = self.index.from_index(index)
        return self.get_scan_by_id(id_str)

    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        if not self._use_index:
            raise TypeError("This method requires the index. Please pass `use_index=True` during initialization")

        scan_ids = tuple(self.index)
        lo = 0
        hi = len(scan_ids)

        best_match = None
        best_error = float('inf')

        if time == float('inf'):
            return self.get_scan_by_id(scan_ids[-1])

        while hi != lo:
            mid = (hi + lo) // 2
            sid = scan_ids[mid]
            scan = self.get_scan_by_id(sid)
            scan_time = scan.scan_time
            err = abs(scan_time - time)
            if err < best_error:
                best_error = err
                best_match = scan
            if scan_time == time:
                return scan
            elif (hi - lo) == 1:
                return best_match
            elif scan_time > time:
                hi = mid
            else:
                lo = mid
        if hi == 0 and not self._use_index:
            raise TypeError("This method requires the index. Please pass `use_index=True` during initialization")

    @property
    def source(self):
        '''The file parser that this reader consumes.
        '''
        return self._source

    @property
    def index(self):
        '''The byte offset index used to achieve fast random access.

        Maps :class:`~.ScanBase` IDs to the byte offsets, implying
        the order the scans reside in the file.

        Returns
        -------
        :class:`pyteomics.xml.ByteEncodingOrderedDict`
        '''
        return self.source.index

    def __len__(self):
        return len(self.index)

    def close(self):
        '''Close the underlying reader.
        '''
        self._source.close()

    def reset(self):
        """Reset the object, clearing out any existing
        state.

        This resets the underlying file iterator, then
        calls :meth:`make_iterator`, and clears the scan
        cache.
        """
        self._source.reset()
        try:
            self.source.seek(0)
        except (IOError, AttributeError):
            pass
        self.make_iterator(None)
        self.initialize_scan_cache()

    def _make_default_iterator(self):
        return iter(self._source)

    def make_iterator(self, iterator=None, grouped=False):
        """Configure the iterator's behavior.

        Parameters
        ----------
        iterator : Iterator, optional
            The iterator to manipulate. If missing, the default
            iterator will be used.
        grouped : bool, optional
            Whether the iterator should be grouped and produce :class:`.ScanBunch` objects
            or single :class:`.Scan`. Defaults to False
        """
        return super(MGFLoader, self).make_iterator(iterator, grouped)

    def _yield_from_index(self, scan_source, start):
        offset_provider = self.index
        keys = list(offset_provider.keys())
        if start is not None:
            if isinstance(start, basestring):
                try:
                    start = keys.index(start)
                except ValueError:
                    start = keys.index(start + '.')
            elif isinstance(start, int):
                start = start
            else:
                raise TypeError("Cannot start from object %r" % start)
        else:
            start = 0
        for key in keys[start:]:
            yield scan_source.get_by_id(key)

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        '''Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        This method will trigger several random-access operations, making it prohibitively
        expensive for normally compressed files.

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        '''
        if scan_id is None:
            if rt is not None:
                scan = self.get_scan_by_time(rt)
            elif index is not None:
                try:
                    scan = self.get_scan_by_index(index)
                except IndexError:
                    if index > len(self.index):
                        index = len(self.index) - 1
                    else:
                        index = 0
                    scan = self.get_scan_by_index(index)

            else:
                raise ValueError("Must provide a scan locator, one of (scan_id, rt, index)")

            scan_id = scan.id
        else:
            scan = self.get_scan_by_id(scan_id)

        # MGF files do not contain MS1 scans
        if require_ms1:
            pass

        iterator = self._yield_from_index(self._source, scan_id)
        self.make_iterator(iterator, grouped=grouped)
        return self

    def _scan_group_iterator(self, iterator=None):
        if iterator is None:
            iterator = self._make_default_iterator()

        impl = _FakeGroupedScanIteratorImpl(
            iterator, self._make_scan, self._validate, self._cache_scan)
        return impl

    def next(self):
        return next(self._producer)

    def _validate(self, scan):
        return True



import os
import logging
import multiprocessing
import traceback

from collections import deque
from multiprocessing import Process
try:
    from Queue import Empty as QueueEmpty
except ImportError:
    from queue import Empty as QueueEmpty

import ms_deisotope

from ms_deisotope.processor import (
    ScanProcessor, MSFileLoader,
    NoIsotopicClustersError, EmptyScanError)

from ms_deisotope.task import show_message


DONE = b"--NO-MORE--"
SCAN_STATUS_GOOD = b"good"
SCAN_STATUS_SKIP = b"skip"


class ScanIDYieldingProcess(Process):

    def __init__(self, ms_file_path, queue, start_scan=None, max_scans=None, end_scan=None,
                 no_more_event=None, ignore_tandem_scans=False, batch_size=1, log_handler=None):
        if log_handler is None:
            log_handler = show_message
        Process.__init__(self)
        self.daemon = True
        self.ms_file_path = ms_file_path
        self.queue = queue
        self.loader = None

        self.start_scan = start_scan
        self.max_scans = max_scans
        self.end_scan = end_scan
        self.ignore_tandem_scans = ignore_tandem_scans
        self.batch_size = batch_size

        self.log_handler = log_handler

        self.no_more_event = no_more_event

    def _make_scan_batch(self):
        batch = []
        scan_ids = []
        for _ in range(self.batch_size):
            try:
                bunch = next(self.loader)
                scan, products = bunch
                if scan is not None:
                    scan_id = scan.id
                else:
                    scan_id = None
                product_scan_ids = [p.id for p in products]
            except StopIteration:
                break
            except Exception as e:
                self.log_handler("An error occurred in _make_scan_batch", e)
                break
            if not self.ignore_tandem_scans:
                batch.append((scan_id, product_scan_ids, True))
            else:
                batch.append((scan_id, product_scan_ids, False))
            scan_ids.append(scan_id)
        return batch, scan_ids

    def run(self):
        self.loader = MSFileLoader(self.ms_file_path)

        if self.start_scan is not None:
            try:
                self.loader.start_from_scan(
                    self.start_scan, require_ms1=self.loader.has_ms1_scans(), grouped=True)
            except IndexError as e:
                self.log_handler("An error occurred while locating start scan", e)
                self.loader.reset()
                self.loader.make_iterator(grouped=True)
            except AttributeError:
                self.log_handler("The reader does not support random access, start time will be ignored", e)
                self.loader.reset()
                self.loader.make_iterator(grouped=True)
        else:
            self.loader.make_iterator(grouped=True)

        count = 0
        last = 0
        if self.max_scans is None:
            max_scans = float('inf')
        else:
            max_scans = self.max_scans

        end_scan = self.end_scan
        while count < max_scans:
            try:
                batch, ids = self._make_scan_batch()
                if len(batch) > 0:
                    self.queue.put(batch)
                count += len(ids)
                if (count - last) > 1000:
                    last = count
                    self.queue.join()
                if (end_scan in ids and end_scan is not None) or len(ids) == 0:
                    self.log_handler("End Scan Found")
                    break
            except StopIteration:
                break
            except Exception as e:
                self.log_handler("An error occurred while fetching scans", e)
                break

        if self.no_more_event is not None:
            self.no_more_event.set()
            self.log_handler("All Scan IDs have been dealt. %d scan bunches." % (count,))
        else:
            self.queue.put(DONE)


class ScanTransformMixin(object):
    def log_error(self, error, scan_id, scan, product_scan_ids):
        tb = traceback.format_exc()
        self.log_handler(
            "An %r occurred for %s (index %r) in Process %r\n%s" % (
                error, scan_id, scan.index, multiprocessing.current_process(),
                tb))

    def _init_batch_store(self):
        self._batch_store = deque()

    def get_work(self, block=True, timeout=30):
        if self._batch_store:
            return self._batch_store.popleft()
        else:
            batch = self.input_queue.get(block, timeout)
            self._batch_store.extend(batch)
            result = self._batch_store.popleft()
            return result

    def log_message(self, message):
        self.log_handler(message + ", %r" %
                         (multiprocessing.current_process().name))

    def skip_entry(self, index, ms_level):
        self.output_queue.put((SCAN_STATUS_SKIP, index, ms_level))

    def skip_scan(self, scan):
        self.output_queue.put((SCAN_STATUS_SKIP, scan.index, scan.ms_level))

    def send_scan(self, scan):
        scan = scan.pack()
        # this attribute is not needed, and for MS1 scans is dangerous
        # to pickle.
        # It can pull other scans which may not yet have been packed
        # into the message sent back to the main process which in
        # turn can form a reference cycle and eat a lot of memory
        scan.product_scans = []
        self.output_queue.put((scan, scan.index, scan.ms_level))

    def all_work_done(self):
        return self._work_complete.is_set()

    def make_scan_transformer(self, loader=None):
        raise NotImplementedError()


class ScanBunchLoader(object):

    def __init__(self, mzml_loader):
        self.loader = mzml_loader
        self.queue = deque()

    def put(self, scan_id, product_scan_ids):
        self.queue.append((scan_id, product_scan_ids))

    def get(self):
        scan_id, product_scan_ids = self.queue.popleft()
        if scan_id is not None:
            precursor = self.loader.get_scan_by_id(scan_id)
        else:
            precursor = None
        products = [self.loader.get_scan_by_id(
            pid) for pid in product_scan_ids if pid is not None]
        if precursor:
            precursor.product_scans = products
        return (precursor, products)


class ScanTransformingProcess(Process, ScanTransformMixin):
    """ScanTransformingProcess describes a child process that consumes scan id bunches
    from a shared input queue, retrieves the relevant scans, and preprocesses them using an
    instance of :class:`ms_deisotope.processor.ScanProcessor`, sending the reduced result
    to a shared output queue.

    Attributes
    ----------
    input_queue : multiprocessing.JoinableQueue
        A shared input queue which contains payloads of bunches of
        scan ids
    ms1_deconvolution_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    ms1_peak_picking_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    msn_deconvolution_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    msn_peak_picking_args : dict
        Parameters passed to :class:`ms_deisotope.processor.ScanProcessor`
    ms_file_path : str
        Path to the spectral data file on disk
    no_more_event : multiprocessing.Event
        An event which will be set when the process feeding the input
        queue has run out of items to add, indicating that any QueueEmptyException
        should be treated as a signal to finish rather than to wait for
        new input
    output_queue : multiprocessing.JoinableQueue
        A shared output queue which this object will put
        :class:`ms_deisotope.data_source.common.ProcessedScan` bunches onto.
    """

    def __init__(self, ms_file_path, input_queue, output_queue,
                 no_more_event=None, ms1_peak_picking_args=None,
                 msn_peak_picking_args=None,
                 ms1_deconvolution_args=None, msn_deconvolution_args=None,
                 envelope_selector=None, ms1_averaging=0, log_handler=None,
                 deconvolute=True, verbose=False, too_many_peaks_threshold=7000):
        if log_handler is None:
            log_handler = show_message

        if ms1_peak_picking_args is None:
            ms1_peak_picking_args = {
                "transforms": [],
                "start_mz": 250
            }
        if msn_peak_picking_args is None:
            msn_peak_picking_args = {
                "transforms": []
            }
        if ms1_deconvolution_args is None:
            ms1_deconvolution_args = {
                "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(20., 2),
                "charge_range": (1, 8),
                "averagine": ms_deisotope.peptide,
                "use_quick_charge": True,
            }
        if msn_deconvolution_args is None:
            msn_deconvolution_args = {
                "scorer": ms_deisotope.scoring.MSDeconVFitter(10.),
                "charge_range": (1, 8),
                "averagine": ms_deisotope.peptide,
                "use_quick_charge": True,
            }

        Process.__init__(self)
        self.daemon = True
        self.verbose = verbose
        self._init_batch_store()
        self.ms_file_path = ms_file_path
        self.input_queue = input_queue
        self.output_queue = output_queue

        self.ms1_peak_picking_args = ms1_peak_picking_args
        self.msn_peak_picking_args = msn_peak_picking_args
        self.ms1_deconvolution_args = ms1_deconvolution_args
        self.msn_deconvolution_args = msn_deconvolution_args
        self.envelope_selector = envelope_selector
        self.ms1_averaging = ms1_averaging
        self.deconvolute = deconvolute

        self.transformer = None

        self.no_more_event = no_more_event
        self._work_complete = multiprocessing.Event()
        self.log_handler = log_handler
        self.too_many_peaks_threshold = too_many_peaks_threshold

    def make_scan_transformer(self, loader=None):
        transformer = ScanProcessor(
            loader,
            ms1_peak_picking_args=self.ms1_peak_picking_args,
            msn_peak_picking_args=self.msn_peak_picking_args,
            ms1_deconvolution_args=self.ms1_deconvolution_args,
            msn_deconvolution_args=self.msn_deconvolution_args,
            loader_type=lambda x: x,
            envelope_selector=self.envelope_selector,
            ms1_averaging=self.ms1_averaging)
        return transformer

    def handle_scan_bunch(self, scan, product_scans, scan_id, product_scan_ids, process_msn=True):
        transformer = self.transformer
        # handle the MS1 scan if it is present
        if scan is not None:
            if len(scan.arrays[0]) == 0:
                self.skip_scan(scan)
                return

            try:
                scan, priorities, product_scans = transformer.process_scan_group(
                    scan, product_scans)
                if scan is None:
                    # no way to report skip
                    pass
                else:
                    if self.verbose:
                        self.log_message("Handling Precursor Scan %r with %d peaks" % (scan.id, len(scan.peak_set)))
                    if self.deconvolute:
                        transformer.deconvolute_precursor_scan(scan, priorities)
                    self.send_scan(scan)
            except NoIsotopicClustersError as e:
                self.log_message("No isotopic clusters were extracted from scan %s (%r)" % (
                    e.scan_id, len(scan.peak_set)))
                self.skip_scan(scan)
            except EmptyScanError as e:
                self.skip_scan(scan)
            except Exception as e:
                self.skip_scan(scan)
                self.log_error(e, scan_id, scan, (product_scan_ids))

        for product_scan in product_scans:
            # no way to report skip
            if product_scan is None:
                continue
            if len(product_scan.arrays[0]) == 0 or (not process_msn):
                self.skip_scan(product_scan)
                continue
            try:
                transformer.pick_product_scan_peaks(product_scan)
                if self.verbose:
                    self.log_message("Handling Product Scan %r with %d peaks (%0.3f/%0.3f, %r)" % (
                        product_scan.id, len(product_scan.peak_set), product_scan.precursor_information.mz,
                        product_scan.precursor_information.extracted_mz,
                        product_scan.precursor_information.defaulted))
                if self.deconvolute:
                    transformer.deconvolute_product_scan(product_scan)
                    if scan is None:
                        product_scan.precursor_information.default(orphan=True)
                self.send_scan(product_scan)
            except NoIsotopicClustersError as e:
                self.log_message("No isotopic clusters were extracted from scan %s (%r)" % (
                    e.scan_id, len(product_scan.peak_set)))
                self.skip_scan(product_scan)
            except EmptyScanError as e:
                self.skip_scan(product_scan)
            except Exception as e:
                self.skip_scan(product_scan)
                self.log_error(e, product_scan.id,
                               product_scan, (product_scan_ids))

    def _silence_loggers(self):
        nologs = ["deconvolution_scan_processor"]
        if not self.deconvolute:
            nologs.append("deconvolution")

        debug_mode = os.getenv("MS_DEISOTOPE_DEBUG")
        if debug_mode:
            handler = logging.FileHandler("ms-deisotope-deconvolution-debug-%s.log" % (os.getpid()), 'w')
            fmt = logging.Formatter(
                "%(asctime)s - %(name)s:%(filename)s:%(lineno)-4d - %(levelname)s - %(message)s",
                "%H:%M:%S")
            handler.setFormatter(fmt)
        for logname in nologs:
            logger_to_silence = logging.getLogger(logname)
            if debug_mode:
                logger_to_silence.setLevel("DEBUG")
                logger_to_silence.addHandler(handler)
            else:
                logger_to_silence.propagate = False
                logger_to_silence.setLevel("CRITICAL")
                logger_to_silence.addHandler(logging.NullHandler())

    def run(self):
        loader = MSFileLoader(self.ms_file_path)
        queued_loader = ScanBunchLoader(loader)

        has_input = True
        transformer = self.make_scan_transformer(loader)
        self.transformer = transformer
        self._silence_loggers()

        i = 0
        last = 0
        while has_input:
            try:
                scan_id, product_scan_ids, process_msn = self.get_work(True, 10)
                self.input_queue.task_done()
            except QueueEmpty:
                if self.no_more_event is not None and self.no_more_event.is_set():
                    has_input = False
                continue

            i += 1 + len(product_scan_ids)
            if scan_id == DONE:
                has_input = False
                break

            try:
                queued_loader.put(scan_id, product_scan_ids)
                scan, product_scans = queued_loader.get()
            except Exception as e:
                self.log_message("Something went wrong when loading bunch (%s): %r.\nRecovery is not possible." % (
                    (scan_id, product_scan_ids), e))

            self.handle_scan_bunch(scan, product_scans, scan_id, product_scan_ids, process_msn)
            if (i - last) > 1000:
                last = i
                self.output_queue.join()

        self.log_message("Done (%d scans)" % i)

        if self.no_more_event is None:
            self.output_queue.put((DONE, DONE, DONE))

        self._work_complete.set()



'''Deconvolution strategies using a list of compositions.
'''

from ms_deisotope.averagine import (
    PROTON, isotopic_variants,
    TheoreticalIsotopicPattern,
    neutral_mass)

from ms_deisotope.envelope_statistics import average_mz, a_to_a2_ratio, most_abundant_mz
from ms_deisotope.constants import (
    IGNORE_BELOW,
    TRUNCATE_AFTER,
    ERROR_TOLERANCE,
    MAX_ITERATION,
    CONVERGENCE)

from ms_deisotope.peak_dependency_network import PeakDependenceGraph
from ms_deisotope.peak_set import DeconvolutedPeakSolution, DeconvolutedPeakSet
from ms_deisotope.scoring import IsotopicFitRecord

from .base import (DeconvoluterBase)


from .utils import (
    count_placeholders, prepare_peaklist,
    drop_placeholders, first_peak, mean, charge_range_,
    )


class CompositionListDeconvoluterBase(DeconvoluterBase):
    """A mixin class to provide common features for deconvoluters which process spectra
    using a list of targeted compositions.

    Attributes
    ----------
    composition_list : list of :class:`~.Mapping`
        A series of objects which represent elemental compositions and support
        the :class:`~.Mapping` interface to access their individual elements.
    """

    def __init__(self, composition_list, *args, **kwargs):
        self.composition_list = list(composition_list)
        self.incremental_truncation = kwargs.get(
            "incremental_truncation", None)
        super(CompositionListDeconvoluterBase, self).__init__(*args, **kwargs)

    def generate_theoretical_isotopic_cluster(self, composition, charge, truncate_after=TRUNCATE_AFTER,
                                              mass_shift=None, charge_carrier=PROTON,
                                              ignore_below=IGNORE_BELOW):
        """Generate a theoretical isotopic pattern for ``composition``

        Parameters
        ----------
        composition : :class:`~.Mapping`
            An object representing an elemental composition
        charge : int
            The charge state to generate the isotopic pattern for
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        mass_shift : float, optional
            An arbitrary mass shift to apply to the generated theoretical isotopic pattern,
            moving all peaks forward by that mass charge ratio transformed mass.
        charge_carrier : float, optional
            The mass of the charge carrier, or more specifically, the moiety which is added for
            each incremental change in charge state. Defaults to |PROTON|

        Returns
        -------
        :class:`~.TheoreticalIsotopicPattern`
            The theoretical isotopic pattern generated
        """
        tid = isotopic_variants(
            composition, charge=charge, charge_carrier=charge_carrier)
        tid = TheoreticalIsotopicPattern(tid, tid[0].mz)
        tid.truncate_after(truncate_after)
        tid.ignore_below(ignore_below)
        if mass_shift is not None:
            tid.shift(mass_shift / abs(charge))
        return tid

    def recalibrate_theoretical_mz(self, theoretical_distribution, experimental_mz):
        """Recalibrate the m/z of the theoretical isotopic pattern to start from the
        peak matching the experimental monoisotopic m/z

        Parameters
        ----------
        theoretical_distribution : :class:`TheoreticalIsotopicPattern`
            The theoretical isotopic pattern to adjust
        experimental_mz : float
            The experimental monoisotopic peak m/z

        Returns
        -------
        TheoreticalIsotopicPattern
        """
        theoretical_distribution.shift(experimental_mz)
        return theoretical_distribution

    def fit_composition_at_charge(self, composition, charge, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON,
                                  truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW, mass_shift=None):
        """Produce an isotopic fit for `composition` at `charge` against the experimental peak set.

        This method requires that the instance also possess a method named `match_theoretical_isotopic_distribution`
        such as the one implemented in :class:`DeconvoluterBase`.

        Parameters
        ----------
        composition : :class:`~.Mapping`
            An object representing an elemental composition
        charge : int
            The charge state to generate the isotopic pattern for
        error_tolerance : float
            The mass accuracy required to for peak matches
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        mass_shift : float, optional
            An arbitrary mass shift to apply to the generated theoretical isotopic pattern,
            moving all peaks forward by that mass charge ratio transformed mass.
        charge_carrier : float, optional
            The mass of the charge carrier, or more specifically, the moiety which is added for
            each incremental change in charge state. Defaults to |PROTON|

        Returns
        -------
        :class:`~.IsotopicFitRecord`
        """
        tid = self.generate_theoretical_isotopic_cluster(composition, charge=charge, truncate_after=truncate_after,
                                                         charge_carrier=charge_carrier, ignore_below=ignore_below,
                                                         mass_shift=mass_shift)
        monoisotopic_peak = self.peaklist.has_peak(tid[0].mz, error_tolerance)
        if monoisotopic_peak is not None:
            tid = self.recalibrate_theoretical_mz(tid, monoisotopic_peak.mz)
        eid = self.match_theoretical_isotopic_distribution(
            tid.peaklist, error_tolerance)

        missed_peaks = count_placeholders(eid)

        if missed_peaks > len(eid) / 2:
            return None

        self.scale_theoretical_distribution(tid, eid)
        score = self.scorer.evaluate(self.peaklist, eid, tid.peaklist)
        fit = IsotopicFitRecord(None, score, charge, tid, eid)
        fit.missed_peaks = missed_peaks
        return fit

    def _make_deconvoluted_peak_solution(self, fit, composition, charge_carrier):
        eid = fit.experimental
        tid = fit.theoretical
        charge = fit.charge
        rep_eid = drop_placeholders(eid)
        total_abundance = sum(
            p.intensity for p in eid if p.intensity > 1)

        monoisotopic_mass = neutral_mass(
            tid.monoisotopic_mz, charge, charge_carrier)
        monoisotopic_mz = tid.monoisotopic_mz

        reference_peak = first_peak(eid)
        peak = DeconvolutedPeakSolution(
            composition, fit,
            monoisotopic_mass, total_abundance, charge,
            signal_to_noise=mean(p.signal_to_noise for p in rep_eid),
            index=reference_peak.index,
            full_width_at_half_max=mean(
                p.full_width_at_half_max for p in rep_eid),
            a_to_a2_ratio=a_to_a2_ratio(tid),
            most_abundant_mass=neutral_mass(
                most_abundant_mz(eid), charge),
            average_mass=neutral_mass(average_mz(eid), charge),
            score=fit.score,
            envelope=[(p.mz, p.intensity) for p in rep_eid],
            mz=monoisotopic_mz, area=sum(e.area for e in eid))
        return peak

    def deconvolute_composition(self, composition, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                                charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW,
                                mass_shift=None):
        """For each charge state under consideration, fit the theoretical isotopic pattern for this composition,
        and if the fit is satisfactory, add it to the results set.

        Parameters
        ----------
        composition : :class:`~.Mapping`
            An object representing an elemental composition
        error_tolerance : float
            The mass accuracy required to for peak matches
        charge_range : tuple
            The charge state range to generate the isotopic patterns for
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        mass_shift : float, optional
            An arbitrary mass shift to apply to the generated theoretical isotopic pattern,
            moving all peaks forward by that mass charge ratio transformed mass.
        charge_carrier : float, optional
            The mass of the charge carrier, or more specifically, the moiety which is added for
            each incremental change in charge state. Defaults to |PROTON|
        """
        for charge in charge_range_(*charge_range):
            fit = self.fit_composition_at_charge(composition, charge=charge, error_tolerance=error_tolerance,
                                                 truncate_after=truncate_after, charge_carrier=charge_carrier,
                                                 mass_shift=mass_shift, ignore_below=ignore_below)
            if fit is None:
                continue
            if not self.scorer.reject(fit):
                eid = fit.experimental
                tid = fit.theoretical
                rep_eid = drop_placeholders(eid)
                if (len(rep_eid) < 2) or (len(rep_eid) < (len(tid) / 2.)) or (len(rep_eid) == 1 and fit.charge > 1):
                    continue
                if self.incremental_truncation is not None:
                    fits = [fit]
                    for case in self.fit_incremental_truncation(fit, self.incremental_truncation):
                        if not self.scorer.reject(case):
                            fits.append(case)
                    fit = self.scorer.select.best(fits)
                peak = self._make_deconvoluted_peak_solution(
                    fit, composition, charge_carrier)
                self._deconvoluted_peaks.append(peak)
                if self.use_subtraction:
                    self.subtraction(tid, error_tolerance)


class CompositionListDeconvoluter(CompositionListDeconvoluterBase):
    '''Fit exact isotopic patterns from a list of compositions.

    Fits are accepted as they are made, making this algorithm unsuitable for
    complex spectra where isotopic patterns will share peaks.

    Attributes
    ----------
    composition_list : list of :class:`~.Mapping`
        A series of objects which represent elemental compositions and support
        the :class:`~.Mapping` interface to access their individual elements.
    peaklist : :class:`~ms_peak_picker.PeakSet`
        The collection of :class:`~.ms_peak_picker.FittedPeak` instances and possible associated
        data to deconvolute.
    scorer : :class:`~.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information
    '''

    def __init__(self, peaklist, composition_list, scorer,
                 use_subtraction=False, scale_method='sum',
                 verbose=False, use_quick_charge=False, **kwargs):
        self.peaklist = prepare_peaklist(peaklist)
        self.scorer = scorer
        self.verbose = verbose
        self._deconvoluted_peaks = []
        self.use_quick_charge = use_quick_charge
        super(CompositionListDeconvoluter, self).__init__(
            composition_list,
            use_subtraction=use_subtraction, scale_method=scale_method,
            merge_isobaric_peaks=True, **kwargs)

    def deconvolute(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), charge_carrier=PROTON,
                    truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW, mass_shift=None, **kwargs):
        """Deconvolute the spectrum, extracting isotopic patterns from the composition list.

        Parameters
        ----------
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        mass_shift: float, optional
            An optional mass shift to apply to each composition

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        for composition in self.composition_list:
            self.deconvolute_composition(composition, error_tolerance=error_tolerance,
                                         charge_range=charge_range, charge_carrier=charge_carrier,
                                         truncate_after=truncate_after, ignore_below=ignore_below,
                                         mass_shift=mass_shift)
        return DeconvolutedPeakSet(self._deconvoluted_peaks).reindex()


class CompositionListPeakDependenceGraphDeconvoluter(CompositionListDeconvoluter):
    '''Fit exact isotopic patterns from a list of compositions.

    Fits are added to a peak dependence graph, and the best fit is chosen after
    all fits are calculated at each iteration.

    Attributes
    ----------
    composition_list : list of :class:`~.Mapping`
        A series of objects which represent elemental compositions and support
        the :class:`~.Mapping` interface to access their individual elements.
    peaklist : :class:`~ms_peak_picker.PeakSet`
        The collection of ms_peak_picker.FittedPeak instances and possible associated
        data to deconvolute.
    scorer : :class:`~.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    max_missed_peaks : int
        The maximum number of missing peaks to tolerate in an isotopic fit
    peak_dependency_network : :class:`~PeakDependenceGraph`
        The peak dependence graph onto which isotopic fit dependences on peaks
        are constructed and solved.
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information
    '''

    def __init__(self, peaklist, composition_list, scorer,
                 use_subtraction=False, scale_method='sum',
                 verbose=False, use_quick_charge=False, **kwargs):
        max_missed_peaks = kwargs.get("max_missed_peaks", 1)
        super(CompositionListPeakDependenceGraphDeconvoluter, self).__init__(
            peaklist, composition_list, scorer=scorer, use_subtraction=use_subtraction,
            scale_method=scale_method, verbose=verbose, use_quick_charge=use_quick_charge,
            **kwargs)

        self.peak_dependency_network = PeakDependenceGraph(
            self.peaklist, maximize=self.scorer.is_maximizing(), **kwargs)
        self.max_missed_peaks = max_missed_peaks

    @property
    def max_missed_peaks(self):
        """The maximum number of missed peaks per isotopic fit record permitted.

        This property directly mirrors :attr:`PeakDependenceGraph.max_missed_peaks`

        Returns
        -------
        int
        """
        return self.peak_dependency_network.max_missed_peaks

    @max_missed_peaks.setter
    def max_missed_peaks(self, value):
        self.peak_dependency_network.max_missed_peaks = value

    def _save_peak_solution(self, solution):
        self._deconvoluted_peaks.append(solution)

    def deconvolute_composition(self, composition, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                                charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW,
                                mass_shift=None):
        for charge in charge_range_(*charge_range):
            fit = self.fit_composition_at_charge(
                composition, charge, error_tolerance, charge_carrier=charge_carrier,
                truncate_after=truncate_after, mass_shift=mass_shift, ignore_below=ignore_below)
            if fit is None:
                continue
            rep_eid = drop_placeholders(fit.experimental)
            if len(rep_eid) == 1 and fit.charge > 1:
                continue
            if not self.scorer.reject(fit):
                self.peak_dependency_network.add_fit_dependence(fit)
            if self.incremental_truncation is not None:
                for case in self.fit_incremental_truncation(fit, self.incremental_truncation):
                    if not self.scorer.reject(case):
                        self.peak_dependency_network.add_fit_dependence(case)


    def populate_graph(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), truncate_after=TRUNCATE_AFTER,
                       charge_carrier=PROTON, ignore_below=IGNORE_BELOW, mass_shift=None):
        """For each composition, for each charge state under consideration, fit the theoretical
        isotopic pattern for this composition, and if the fit is satisfactory, add it to the
        peak dependence graph for later selecting the optimal solution.

        Parameters
        ----------
        error_tolerance : float
            The mass accuracy required to for peak matches
        charge_range : tuple
            The charge state range to generate the isotopic patterns for
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        charge_carrier : float, optional
            The mass of the charge carrier, or more specifically, the moiety which is added for
            each incremental change in charge state. Defaults to |PROTON|
        mass_shift : float, optional
            An arbitrary mass shift to apply to the generated theoretical isotopic pattern,
            moving all peaks forward by that mass charge ratio transformed mass.
        """
        for composition in self.composition_list:
            self.deconvolute_composition(composition, error_tolerance, charge_range,
                                         truncate_after=truncate_after, charge_carrier=charge_carrier,
                                         ignore_below=ignore_below, mass_shift=mass_shift)

    def select_best_disjoint_subgraphs(self, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON):
        """Construct connected envelope graphs from :attr:`peak_dependency_network` and
        extract the best disjoint isotopic pattern fits in each envelope graph. This in
        turn produces one or more :class:`~.DeconvolutedPeakSolution` instances from each
        disjoint fit, which are processed and added to the results set.

        Parameters
        ----------
        error_tolerance : float, optional
            The error tolerance to use when performing subtraction, if subtraction is
            being performed.
        charge_carrier : float, optional
            The mass of the charge carrier as used for the deconvolution. Required to
            back-out the neutral mass of the deconvoluted result
        """
        disjoint_envelopes = self.peak_dependency_network.find_non_overlapping_intervals()

        for cluster in disjoint_envelopes:
            for fit in cluster.disjoint_best_fits():
                eid = fit.experimental
                tid = fit.theoretical
                composition = fit.data
                rep_eid = drop_placeholders(eid)
                if len(rep_eid) < 2 or len(rep_eid) < len(tid) / 2.:
                    continue

                peak = self._make_deconvoluted_peak_solution(
                    fit, composition, charge_carrier)
                self._save_peak_solution(peak)
                if self.use_subtraction:
                    self.subtraction(tid, error_tolerance)

    def deconvolute(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), iterations=MAX_ITERATION,   # pylint: disable=arguments-differ
                    truncate_after=TRUNCATE_AFTER, charge_carrier=PROTON, ignore_below=IGNORE_BELOW,
                    mass_shift=None, convergence=CONVERGENCE, **kwargs):
        """Deconvolute the spectrum, extracting isotopic patterns from the composition list.

        Parameters
        ----------
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        mass_shift: float, optional
            An optional mass shift to apply to each composition
        convergence : float, optional
            The threshold of the below which after the `(sum(intensity_before) - sum(
            intensity_after)) / sum(intensity_after)`

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        if not self.use_subtraction:
            iterations = 1
        begin_signal = sum([p.intensity for p in self.peaklist])
        for _ in range(iterations):
            self.populate_graph(error_tolerance, charge_range, charge_carrier=charge_carrier,
                                truncate_after=truncate_after, ignore_below=ignore_below,
                                mass_shift=mass_shift)
            self.select_best_disjoint_subgraphs(error_tolerance)
            self._slice_cache.clear()
            end_signal = sum([p.intensity for p in self.peaklist]) + 1
            if (begin_signal - end_signal) / end_signal < convergence:
                break
            begin_signal = end_signal
        return DeconvolutedPeakSet(self._deconvoluted_peaks).reindex()



from .common import (
    ScanSerializerBase, ScanDeserializerBase)

from .mzml import (
    ProcessedMzMLDeserializer, MzMLSerializer)

from .text import (
    TextScanSerializerBase, HeaderedDelimitedWriter)

from .mgf import (
    ProcessedMGFDeserializer, MGFSerializer)


__all__ = [
    "ScanSerializerBase",
    "ScanDeserializerBase",
    "MzMLSerializer",
    "ProcessedMzMLDeserializer",
    "MGFSerializer",
    "ProcessedMGFDeserializer",
    "TextScanSerializerBase",
    "HeaderedDelimitedWriter"
]



# cython: embedsignature=True

cimport cython
from cpython.list cimport PyList_Append, PyList_GET_ITEM, PyList_GET_SIZE

import numpy as np
cimport numpy as np

from ms_peak_picker._c.peak_index cimport PeakIndex
from ms_peak_picker._c.peak_set cimport PeakSet
from ms_deisotope._c.peak_set cimport (
    Envelope, EnvelopePair, DeconvolutedPeak, DeconvolutedPeakSet, DeconvolutedPeakSetIndexed, PeakBase)
from ms_deisotope._c.averagine cimport neutral_mass


np.import_array()


@cython.cdivision(True)
cdef double ppm_error(double x, double y):
    return (x - y) / y


cdef double ppm2da(double mass, double error_tolerance):
    x = mass + mass * error_tolerance
    return x - mass


@cython.cdivision(True)
cdef double da2ppm(double mass, double error_tolerance):
    x = mass + error_tolerance
    return (x - mass) / mass


cpdef intensity_getter(peak):
    return (<PeakBase>peak).intensity


cpdef mz_getter(peak):
    return (<PeakBase>peak).mz


cpdef mass_getter(peak):
    return (<DeconvolutedPeak>peak).neutral_mass



@cython.boundscheck(False)
cpdef list decode_envelopes(np.ndarray[np.float32_t, ndim=1] array):
    cdef:
        list envelope_list, current_envelope
        tuple current_envelope_tuple
        size_t i, n, n_members
        np.float32_t a, b
    envelope_list = []
    current_envelope = []
    i = 0
    n = array.shape[0]
    n_members = 0
    while i < n:
        a = array[i]
        b = array[i + 1]
        i += 2
        if a == 0 and b == 0:
            if current_envelope is not None:
                if n_members > 0:
                    current_envelope_tuple = tuple(current_envelope)
                    PyList_Append(envelope_list, Envelope._create(current_envelope_tuple))
                current_envelope = []
                n_members = 0
        else:
            PyList_Append(current_envelope, EnvelopePair._create(a, b))
            n_members += 1
    current_envelope_tuple = tuple(current_envelope)
    envelope_list.append(Envelope._create(current_envelope_tuple))
    return envelope_list


@cython.boundscheck(False)
cpdef DeconvolutedPeakSetIndexed deserialize_deconvoluted_peak_set(dict scan_dict):
    cdef:
        list envelopes, peaks
        np.ndarray[np.float32_t] mz_array
        np.ndarray[np.float32_t] intensity_array
        np.ndarray[np.int8_t] charge_array
        np.ndarray[np.float32_t] score_array
        size_t n, i
        double mz, peak_neutral_mass
        int charge
        DeconvolutedPeak peak
        DeconvolutedPeakSetIndexed peak_set

    peaks = []
    envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
    mz_array = scan_dict['m/z array'].astype(np.float32)
    intensity_array = scan_dict['intensity array'].astype(np.float32)
    charge_array = scan_dict['charge array'].astype(np.int8)
    score_array = scan_dict['deconvolution score array'].astype(np.float32)
    n = mz_array.shape[0]
    i = 0
    for i in range(n):
        mz = mz_array[i]
        charge = charge_array[i]
        peak_neutral_mass = neutral_mass(mz, charge)

        peak = DeconvolutedPeak._create_simple(
            peak_neutral_mass,
            intensity_array[i],
            charge_array[i],
            score_array[i],
            mz,
            <Envelope>PyList_GET_ITEM(envelopes, i))
        peaks.append(peak)
    peak_set = DeconvolutedPeakSetIndexed(peaks)
    peak_set.reindex()
    return peak_set


@cython.boundscheck(False)
cpdef DeconvolutedPeakSetIndexed build_deconvoluted_peak_set_from_arrays(np.ndarray[double, ndim=1] mz_array,
                                                                         np.ndarray[double, ndim=1] intensity_array,
                                                                         np.ndarray[long, ndim=1] charge_array):
    cdef:
        list peaks
        size_t n, i
        double mz, peak_neutral_mass
        int charge
        DeconvolutedPeak peak
        DeconvolutedPeakSetIndexed peak_set

    peaks = []
    n = mz_array.shape[0]
    i = 0
    for i in range(n):
        mz = mz_array[i]
        charge = charge_array[i]
        peak_neutral_mass = neutral_mass(mz, charge)

        peak = DeconvolutedPeak._create_simple(
            peak_neutral_mass,
            intensity_array[i],
            charge_array[i],
            intensity_array[i],
            mz,
            None)
        peaks.append(peak)
    peak_set = DeconvolutedPeakSetIndexed(peaks)
    peak_set.reindex()
    return peak_set


@cython.binding(True)
cpdef double _peak_sequence_tic(self, peak_collection peaks) except -1:
    cdef:
        size_t i, n
        PeakBase peak
        double tic
        list py_peaks

    tic = 0.0
    if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet:
        n = peaks.get_size()
    else:
        py_peaks = list(peaks)
        n = PyList_GET_SIZE(py_peaks)
        if n > 0:
            if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
                raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
    for i in range(n):
        if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet:
            peak = peaks.getitem(i)
        elif peak_collection is PeakIndex:
            peak = peaks.peaks.getitem(i)
        else:
            peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
        tic += peak.intensity
    return tic


@cython.binding(True)
cpdef PeakBase _peak_sequence_bp(self, peak_collection peaks):
    cdef:
        size_t i, n
        PeakBase peak, base_peak
        double max_intensity
        list py_peaks

    max_intensity = 0.0
    base_peak = None
    if peak_collection is PeakSet or peak_collection is PeakIndex or peak_collection is DeconvolutedPeakSet:
        n = peaks.get_size()
    else:
        py_peaks = list(peaks)
        n = PyList_GET_SIZE(py_peaks)
        if n > 0:
            if not isinstance(<object>PyList_GET_ITEM(py_peaks, 0), PeakBase):
                raise TypeError("Cannot interpret %r as a PeakBase object" % (type(py_peaks[0])))
    for i in range(n):
        if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet:
            peak = peaks.getitem(i)
        elif peak_collection is PeakIndex:
            peak = peaks.peaks.getitem(i)
        else:
            peak = <PeakBase>PyList_GET_ITEM(py_peaks, i)
        if max_intensity < peak.intensity:
            base_peak = peak
            max_intensity = peak.intensity
    return base_peak



from collections import OrderedDict, namedtuple

import numpy as np
from scipy.optimize import leastsq
from numpy import pi, sqrt, exp
from scipy.special import erf

from ms_peak_picker import search

from .profile_transform import smooth_leveled, ValleyPoint, PeakBoundary, ProfileSplitter


DEFAULT_SMOOTH = 3


def linear_regression_residuals(x, y):
    X = np.vstack((np.ones(len(x)), np.array(x))).T
    Y = np.array(y)
    B = np.linalg.inv(X.T.dot(X)).dot(X.T.dot(Y))
    Yhat = X.dot(B)
    return (Y - Yhat) ** 2


class PeakShapeModelBase(object):
    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)


class SkewedGaussianModel(PeakShapeModelBase):
    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma, gamma = params
        return ys - SkewedGaussianModel.shape(xs, center, amplitude, sigma, gamma) * (
            sigma / 2. if abs(sigma) > 2 else 1.)

    @staticmethod
    def guess(xs, ys):
        center = np.average(xs, weights=ys / ys.sum())
        height_at = np.abs(xs - center).argmin()
        apex = ys[height_at]
        sigma = np.abs(center - xs[[search.nearest_left(ys, apex / 2, height_at),
                                    search.nearest_right(ys, apex / 2, height_at + 1)]]).sum()
        gamma = 1
        return center, apex, sigma, gamma

    @staticmethod
    def params_to_dict(params):
        center, amplitude, sigma, gamma = params
        return OrderedDict((("center", center), ("amplitude", amplitude), ("sigma", sigma), ("gamma", gamma)))

    @staticmethod
    def shape(xs, center, amplitude, sigma, gamma):
        norm = (amplitude) / (sigma * sqrt(2 * pi)) * \
            exp(-((xs - center) ** 2) / (2 * sigma ** 2))
        skew = (1 + erf((gamma * (xs - center)) / (sigma * sqrt(2))))
        return norm * skew

    @staticmethod
    def center(params_dict):
        return params_dict['center']

    @staticmethod
    def spread(params_dict):
        return params_dict['sigma']


class PenalizedSkewedGaussianModel(SkewedGaussianModel):
    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma, gamma = params
        return ys - PenalizedSkewedGaussianModel.shape(xs, center, amplitude, sigma, gamma) * (
            sigma / 2. if abs(sigma) > 2 else 1.) * (gamma / 2. if abs(gamma) > 40 else 1.) * (
            center if center > xs[-1] or center < xs[0] else 1.)


class BiGaussianModel(PeakShapeModelBase):

    @staticmethod
    def center(params_dict):
        return params_dict['center']

    @staticmethod
    def spread(params_dict):
        return (params_dict['sigma_left'] + params_dict['sigma_right']) / 2.

    @staticmethod
    def shape(xs, center, amplitude, sigma_left, sigma_right):
        ys = np.zeros_like(xs, dtype=np.float32)
        left_mask = xs < center
        ys[left_mask] = amplitude * np.exp(-(xs[left_mask] - center) ** 2 / (2 * sigma_left ** 2)) * sqrt(2 * pi)
        right_mask = xs > center
        ys[right_mask] = amplitude * np.exp(-(xs[right_mask] - center) ** 2 / (2 * sigma_right ** 2)) * sqrt(2 * pi)
        return ys

    @staticmethod
    def fit(params, xs, ys):
        center, amplitude, sigma_left, sigma_right = params
        return ys - BiGaussianModel.shape(
            xs, center, amplitude, sigma_left, sigma_right) * (
            center if center > xs[-1] or center < xs[0] else 1.)

    @staticmethod
    def params_to_dict(params):
        center, amplitude, sigma_left, sigma_right = params
        return OrderedDict(
            (("center", center), ("amplitude", amplitude), ("sigma_left", sigma_left), ("sigma_right", sigma_right)))

    @staticmethod
    def guess(xs, ys):
        center = np.average(xs, weights=ys / ys.sum())
        height_at = np.abs(xs - center).argmin()
        apex = ys[height_at]
        sigma = np.abs(center - xs[[search.nearest_left(ys, apex / 2, height_at),
                                    search.nearest_right(ys, apex / 2, height_at + 1)]]).sum()
        return center, apex, sigma, sigma


class FittedPeakShape(object):
    def __init__(self, params, shape_model):
        self.params = params
        self.shape_model = shape_model

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def items(self):
        return self.params.items()

    def __iter__(self):
        return iter(self.params)

    def shape(self, xs):
        return self.shape_model.shape(xs, **self.params)

    def __getitem__(self, key):
        return self.params[key]

    def __repr__(self):
        return "Fitted{self.shape_model.__class__.__name__}({params})".format(
            self=self, params=", ".join("%s=%0.3f" % (k, v) for k, v in self.params.items()))

    @property
    def center(self):
        return self['center']

    @property
    def amplitude(self):
        return self['amplitude']


class ChromatogramShapeFitterBase(object):
    def __init__(self, chromatogram, smooth=DEFAULT_SMOOTH, fitter=PenalizedSkewedGaussianModel()):
        self.chromatogram = chromatogram
        self.smooth = smooth
        self.xs = None
        self.ys = None
        self.line_test = None
        self.off_center = None
        self.shape_fitter = fitter

    def handle_invalid(self):
        self.line_test = 0.5

    def extract_arrays(self):
        self.xs, self.ys = self.chromatogram.as_arrays()
        if self.smooth:
            self.ys = smooth_leveled(self.xs, self.ys, self.smooth)
        if len(self.xs) > 2000:
            new_xs = np.linspace(self.xs.min(), self.xs.max(), 2000)
            new_ys = np.interp(new_xs, self.xs, self.ys)
            self.xs = new_xs
            self.ys = new_ys
            self.ys = smooth_leveled(self.xs, self.ys, self.smooth)

    def compute_residuals(self):
        return NotImplemented

    def perform_line_test(self):
        residuals = self.compute_residuals()
        null_residuals = linear_regression_residuals(self.xs, self.ys)
        line_test = (residuals ** 2).sum() / (null_residuals).sum()
        if line_test > 1.0:
            line_test = 1.0
        self.line_test = line_test

    def plot(self, ax=None):
        if ax is None:
            from matplotlib import pyplot as plt
            fig, ax = plt.subplots(1)
        ax.plot(self.xs, self.ys, label='Observed')
        ax.scatter(self.xs, self.ys, label='Observed')
        ax.plot(self.xs, self.compute_fitted(), label='Fitted')
        ax.plot(self.xs, self.compute_residuals(), label='Residuals')
        return ax

    @property
    def fit_parameters(self):
        raise NotImplementedError()


class ChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, smooth=DEFAULT_SMOOTH, fitter=PenalizedSkewedGaussianModel()):
        super(ChromatogramShapeFitter, self).__init__(chromatogram, smooth=smooth, fitter=fitter)

        self.params = None
        self.params_dict = None

        if len(chromatogram) < 5:
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()
            self.off_center_factor()

    @property
    def fit_parameters(self):
        return [self.params_dict]

    def __repr__(self):
        return "ChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def off_center_factor(self):
        center = self.shape_fitter.center(self.params_dict)
        spread = self.shape_fitter.spread(self.params_dict)
        self.off_center = abs(1 - abs(1 - (2 * abs(
            self.xs[self.ys.argmax()] - center) / abs(spread))))
        if self.off_center > 1:
            self.off_center = 1. / self.off_center
        self.line_test /= self.off_center

    def compute_residuals(self):
        return self.shape_fitter.fit(self.params, self.xs, self.ys)

    def compute_fitted(self):
        return self.shape_fitter.shape(self.xs, **self.params_dict)

    def peak_shape_fit(self):
        xs, ys = self.xs, self.ys
        params = self.shape_fitter.guess(xs, ys)
        fit = leastsq(self.shape_fitter.fit,
                      params, (xs, ys))
        params = fit[0]
        self.params = params
        self.params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)

    def iterfits(self):
        yield self.compute_fitted()


def shape_fit_test(chromatogram, smooth=DEFAULT_SMOOTH):
    return ChromatogramShapeFitter(chromatogram, smooth).line_test


def peak_indices(x, min_height=0):
    """Find the index of local maxima.

    Parameters
    ----------
    x : np.ndarray
        Data to find local maxima in
    min_height : float, optional
        Minimum peak height

    Returns
    -------
    np.ndarray[int]
        Indices of maxima in x

    References
    ----------
    https://github.com/demotu/BMC/blob/master/functions/detect_peaks.py
    """
    if x.size < 3:
        return np.array([], dtype=int)
    # find indices of all peaks
    dx = x[1:] - x[:-1]
    rising_edges = np.where((np.hstack((dx, 0)) <= 0) &
                            (np.hstack((0, dx)) > 0))[0]
    falling_edges = np.where((np.hstack((dx, 0)) < 0) &
                             (np.hstack((0, dx)) >= 0))[0]
    indices = np.unique(np.hstack((rising_edges, falling_edges)))
    if indices.size and min_height > 0:
        indices = indices[x[indices] >= min_height]
    return indices


class MultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_peaks=5, smooth=DEFAULT_SMOOTH, fitter=BiGaussianModel()):
        super(MultimodalChromatogramShapeFitter, self).__init__(chromatogram, smooth=smooth, fitter=fitter)
        self.max_peaks = max_peaks
        self.params_list = []
        self.params_dict_list = []

        if len(self.chromatogram) < 5:
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()

    @property
    def fit_parameters(self):
        return self.params_dict_list

    def __repr__(self):
        return "MultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def peak_shape_fit(self):
        return self.set_up_peak_fit()

    def set_up_peak_fit(self, ys=None, min_height=0, peak_count=0):
        xs = self.xs
        if ys is None:
            ys = self.ys
        params = self.shape_fitter.guess(xs, ys)
        params_dict = self.shape_fitter.params_to_dict(params)

        indices = peak_indices(ys, min_height)
        if len(indices) > 0:
            center = xs[max(indices, key=lambda x: ys[x])]
        else:
            center = xs[len(xs) / 2]
        params_dict['center'] = center

        fit = leastsq(self.shape_fitter.fit,
                      params_dict.values(), (xs, ys))
        params = fit[0]
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        self.params_list.append(params)
        self.params_dict_list.append(params_dict)

        residuals = self.shape_fitter.fit(params, xs, ys)

        fitted_apex_index = search.get_nearest(xs, params_dict['center'], 0)
        fitted_apex = ys[fitted_apex_index]

        new_min_height = fitted_apex * 0.5

        if new_min_height < min_height:
            min_height *= 0.85
        else:
            min_height = new_min_height

        indices = peak_indices(residuals, min_height)

        peak_count += 1
        if indices.size and peak_count < self.max_peaks:
            residuals, params_dict = self.set_up_peak_fit(residuals, min_height, peak_count=peak_count)

        return residuals, params_dict

    def compute_fitted(self):
        xs = self.xs
        fitted = np.zeros_like(xs)
        for params_dict in self.params_dict_list:
            fitted += self.shape_fitter.shape(xs, **params_dict)
        return fitted

    def compute_residuals(self):
        return self.ys - self.compute_fitted()

    def iterfits(self):
        xs = self.xs
        for params_dict in self.params_dict_list:
            yield self.shape_fitter.shape(xs, **params_dict)


class AdaptiveMultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_peaks=5, smooth=DEFAULT_SMOOTH, fitters=None):
        if fitters is None:
            fitters = (BiGaussianModel(), PenalizedSkewedGaussianModel(),)
        super(AdaptiveMultimodalChromatogramShapeFitter, self).__init__(
            chromatogram, smooth=smooth, fitter=fitters[0])
        self.max_peaks = max_peaks
        self.fitters = fitters
        self.params_list = []
        self.params_dict_list = []

        self.alternative_fits = []
        self.best_fit = None

        if len(self.chromatogram) < 5:
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()

    @property
    def fit_parameters(self):
        return self.params_dict_list

    def compute_fitted(self):
        return self.best_fit.compute_fitted()

    def compute_residuals(self):
        return self.best_fit.compute_residuals()

    def peak_shape_fit(self):
        for fitter in self.fitters:
            model_fit = ProfileSplittingMultimodalChromatogramShapeFitter(
                self.chromatogram, self.max_peaks, self.smooth, fitter=fitter)
            self.alternative_fits.append(model_fit)
            model_fit = MultimodalChromatogramShapeFitter(
                self.chromatogram, self.max_peaks, self.smooth, fitter=fitter)
            self.alternative_fits.append(model_fit)
        self.best_fit = min(self.alternative_fits, key=lambda x: x.line_test)
        self.params_list = self.best_fit.params_list
        self.params_dict_list = self.best_fit.params_dict_list
        self.shape_fitter = self.best_fit.shape_fitter

    def perform_line_test(self):
        self.line_test = self.best_fit.line_test

    def iterfits(self):
        xs = self.xs
        for params_dict in self.params_dict_list:
            yield self.shape_fitter.shape(xs, **params_dict)

    def __repr__(self):
        return "AdaptiveMultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)


# class SplittingPoint(object):
#     __slots__ = ["first_maximum", "minimum", "second_maximum", "minimum_index", "total_distance"]

#     def __init__(self, first_maximum, minimum, second_maximum, minimum_index):
#         self.first_maximum = first_maximum
#         self.minimum = minimum
#         self.second_maximum = second_maximum
#         self.minimum_index = minimum_index
#         self.total_distance = self.compute_distance()

#     def compute_distance(self):
#         return (self.first_maximum - self.minimum) + (self.second_maximum - self.minimum)

#     def __repr__(self):
#         return "SplittingPoint(%0.4f, %0.4f, %0.4f, %0.2f, %0.3e)" % (
#             self.first_maximum, self.minimum, self.second_maximum, self.minimum_index, self.total_distance)


class ProfileSplittingMultimodalChromatogramShapeFitter(ChromatogramShapeFitterBase):
    def __init__(self, chromatogram, max_splits=3, smooth=DEFAULT_SMOOTH, fitter=BiGaussianModel()):
        super(ProfileSplittingMultimodalChromatogramShapeFitter, self).__init__(
            chromatogram, smooth=smooth, fitter=fitter)
        self.max_splits = max_splits
        self.params_list = []
        self.params_dict_list = []
        self.partition_sites = []

        if len(self.chromatogram) < 5:
            self.handle_invalid()
        else:
            self.extract_arrays()
            self.peak_shape_fit()
            self.perform_line_test()

    def __repr__(self):
        return "ProfileSplittingMultimodalChromatogramShapeFitter(%s, %0.4f)" % (self.chromatogram, self.line_test)

    def _extreme_indices(self, ys):
        maxima_indices = peak_indices(ys)
        minima_indices = peak_indices(-ys)
        return maxima_indices, minima_indices

    def locate_extrema(self, xs=None, ys=None):
        if xs is None:
            xs = self.xs
        if ys is None:
            ys = self.ys

        splitter = ProfileSplitter((xs, ys))
        candidates = splitter.locate_valleys(scale=0.01, smooth=self.smooth, interpolate_past=2000)
        if candidates:
            best_point = candidates[0]
            self.partition_sites.append(best_point)

        return candidates

    def build_partitions(self):
        segments = []

        last_x = self.xs.min() - 1
        for point in self.partition_sites:
            mask = (self.xs <= point.minimum_index) & (self.xs > last_x)
            if any(mask):
                segments.append((self.xs[mask], self.ys[mask]))
            last_x = point.minimum_index
        mask = self.xs > last_x
        if any(mask):
            segments.append((self.xs[mask], self.ys[mask]))
        return segments

    def set_up_peak_fit(self, xs, ys):
        params = self.shape_fitter.guess(xs, ys)
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        if len(params) > len(xs):
            self.params_list.append(params)
            self.params_dict_list.append(params_dict)
            return ys, params_dict

        fit = leastsq(self.shape_fitter.fit,
                      params_dict.values(), (xs, ys))
        params = fit[0]
        params_dict = FittedPeakShape(self.shape_fitter.params_to_dict(params), self.shape_fitter)
        self.params_list.append(params)
        self.params_dict_list.append(params_dict)

        residuals = self.shape_fitter.fit(params, xs, ys)
        return residuals, params_dict

    def peak_shape_fit(self):
        self.locate_extrema()
        for segment in self.build_partitions():
            self.set_up_peak_fit(*segment)

    def compute_fitted(self):
        fitted = []
        for segment, params_dict in zip(self.build_partitions(), self.params_dict_list):
            fitted.append(self.shape_fitter.shape(segment[0], **params_dict))
        return np.concatenate(fitted)

    def compute_residuals(self):
        return self.ys - self.compute_fitted()

    def iterfits(self):
        for segment, params_dict in zip(self.build_partitions(), self.params_dict_list):
            yield self.shape_fitter.shape(segment[0], **params_dict)


class CentroidFit(object):
    def __init__(self, center, weight, fits=None):
        if fits is None:
            fits = []
        self.center = center
        self.weight = weight
        self.fits = fits

    def __repr__(self):
        return "CentroidFit(%f, %0.3e, %d)" % (self.center, self.weight, len(self.fits))

    def add(self, fit_params):
        self.fits.append(fit_params)
        self.center, self.weight = self.reweight_center()

    def reweight_center(self):
        center_acc = 0
        weight_acc = 0
        for fit in self.fits:
            center_acc = fit['center'] * fit['amplitude']
            weight_acc = fit['amplitude']
        return center_acc / weight_acc, weight_acc

    @classmethod
    def fromparams(cls, params):
        center = params['center']
        weight = params['amplitude']
        return cls(center, weight, [params])


class ProfileSet(object):
    def __init__(self, features):
        self.features = list(features)
        self.fits = list(map(AdaptiveMultimodalChromatogramShapeFitter, self.features))
        self.baselines = self.compute_baselines()
        self.binned_fits = self.overlap_apexes()

    def compute_baselines(self):
        baselines = []
        for fit in self.fits:
            baselines.append(fit.ys[fit.ys < fit.ys.mean()].mean())
        return baselines

    def overlap_apexes(self, error_tolerance=2e-3):
        centroid_fits = []
        for fit in self.fits:
            centroid_fits.extend(fit.params_dict_list)

        centroid_fits.sort(key=lambda x: x['center'])

        binned_fits = []
        last_fit = CentroidFit.fromparams(centroid_fits[0])
        for fit in centroid_fits[1:]:
            if abs((last_fit.center - fit['center']) / fit['center']) < error_tolerance:
                last_fit.add(fit)
            else:
                binned_fits.append(last_fit)
                last_fit = CentroidFit.fromparams(fit)
        binned_fits.append(last_fit)
        binned_fits.sort(key=lambda x: x.weight, reverse=True)
        return binned_fits

    @staticmethod
    def find_right_intersect(vec, target_val, start_index=0):
        nearest_index = start_index
        next_index = start_index

        size = len(vec) - 1
        if next_index == size:
            return size

        next_val = vec[next_index]
        best_distance = np.abs(next_val - target_val)
        while (next_index < size):
            next_index += 1
            next_val = vec[next_index]
            dist = np.fabs(next_val - target_val)
            if dist < best_distance:
                best_distance = dist
                nearest_index = next_index
            if next_index == size or next_val < target_val:
                break
        return nearest_index

    @staticmethod
    def find_left_intersect(vec, target_val, start_index=0):
        nearest_index = start_index
        next_index = start_index

        size = len(vec) - 1
        if next_index == size:
            return size

        next_val = vec[next_index]
        best_distance = np.abs(next_val - target_val)
        while (next_index > 0):
            next_index -= 1
            next_val = vec[next_index]
            dist = np.fabs(next_val - target_val)
            if dist < best_distance:
                best_distance = dist
                nearest_index = next_index
            if next_index == size or next_val < target_val:
                break
        return nearest_index

    def find_intersects(self, fit_bin=0):
        starts = []
        ends = []
        bin_fit = self.binned_fits[fit_bin]
        for i, feat_fit in enumerate(self.fits):
            baseline = self.baselines[i]
            xs = feat_fit.xs
            ys = feat_fit.compute_fitted()
            center_ix = search.get_nearest(xs, bin_fit.center, 0)
            left_ix = self.find_left_intersect(ys, baseline, center_ix)
            right_ix = self.find_right_intersect(ys, baseline, center_ix)
            starts.append(xs[left_ix])
            ends.append(xs[right_ix])
        return starts, ends

    def find_bounds(self, fit_bin=0):
        starts, ends = self.find_intersects(fit_bin)
        start_acc = 0
        end_acc = 0
        weight = 0
        for i, f in enumerate(self.features):
            start_acc += starts[i] * f.intensity
            end_acc += ends[i] * f.intensity
            weight += f.intensity
        return start_acc / weight, end_acc / weight

    def split(self, fit_bin=0):
        start, end = self.find_bounds(fit_bin)
        before = []
        apex = []
        after = []
        for feat in self.features:
            before_part, rest = feat.split_at(start)
            if len(before_part) > 0:
                before.append(before_part)
            apex_part, after_part = rest.split_at(end)
            if len(apex_part) > 0:
                apex.append(apex_part)
            if len(after_part) > 0:
                after.append(after_part)
        return before, apex, after



import unittest
import platform

import numpy as np

from ms_deisotope.data_source.thermo_raw import (
    determine_if_available)
from ms_deisotope.data_source.thermo_raw_net import (
    determine_if_available as determine_if_available_net)
from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type

not_windows = not platform.platform().lower().startswith("windows")
missing_reader_dll = not determine_if_available()
missing_net_dll = not determine_if_available_net()


class ThermoRawLoaderScanBehaviorBase(object):
    path = datafile("small.RAW")

    reference_mzml = datafile("small.mzML")
    reference_mgf = datafile("small.mgf")

    @property
    def reader(self):  # pragma: no cover
        # Required for implementation
        raise NotImplementedError()

    def test_iteration(self):
        reader = self.reader
        reader.start_from_scan('controllerType=0 controllerNumber=1 scan=10')
        bunch = next(reader)
        assert bunch.precursor.id == 'controllerType=0 controllerNumber=1 scan=9'
        bunch = next(reader)
        assert bunch.precursor.id == 'controllerType=0 controllerNumber=1 scan=15'
        reader.start_from_scan(rt=0.077788333333)
        bunch = next(reader)
        assert np.isclose(bunch.precursor.scan_time, 0.077788333333)

    def test_file_level_metadata(self):
        reader = self.reader
        desc = reader.file_description()
        assert desc.has_content("MS1 spectrum")
        assert desc.has_content("MSn spectrum")

        inst_config = reader.instrument_configuration()
        assert inst_config[0].analyzers[0] == 'orbitrap'

    def test_scan_level_data(self):
        reader = self.reader
        reader.start_from_scan('controllerType=0 controllerNumber=1 scan=10')
        bunch = next(reader)
        assert np.isclose(bunch.precursor.scan_time, 0.077788333333)
        assert len(bunch.precursor.pick_peaks(signal_to_noise_threshold=1.5).peak_set) == 3110
        scan_window = bunch.precursor.acquisition_information.scan_list[0][0]
        assert scan_window.lower == 200.0 and scan_window.upper == 2000.0
        product = bunch.products[0]
        assert product.ms_level == 2
        assert product.index == 9
        assert product.activation.energy == 35.0
        assert np.isclose(product.precursor_information.mz, 810.7528)
        annotations = {'[Thermo Trailer Extra]Micro Scan Count': 3.0,
                       '[Thermo Trailer Extra]Scan Event': 3.0,
                       '[Thermo Trailer Extra]Scan Segment': 1.0,
                       'filter string': 'ITMS + c ESI d Full ms2 810.75@cid35.00 [210.00-1635.00]'}
        assert product.annotations == annotations
        assert np.isclose(product.isolation_window.target, 810.752807)
        assert product.isolation_window.lower == 1.0
        assert not product.is_profile

    def test_size(self):
        reader = self.reader
        n = len(reader)
        assert n == 48
        x = reader[-1]
        y = reader.get_scan_by_time(float('inf'))
        assert x == y

    def test_compat(self):
        raw_reader = self.reader
        mzml_reader = infer_type.MSFileLoader(self.reference_mzml)
        mgf_reader = infer_type.MSFileLoader(self.reference_mgf)

        mgf_scan = next(mgf_reader)
        mzml_scan = mzml_reader[2]
        raw_scan = raw_reader[2]

        self.assertEqual(mzml_scan, raw_scan)

        mgf_scan.pick_peaks()
        raw_scan.pick_peaks()
        self.assertEqual(raw_scan.peak_set, mgf_scan.peak_set)

        self.assertEqual(raw_scan.precursor_information.precursor,
                         mzml_scan.precursor_information.precursor)


@unittest.skipIf(not_windows or missing_reader_dll, "Requires Windows COM and MSFileReader.dll")
class TestCOMThermoRawLoaderScanBehavior(ThermoRawLoaderScanBehaviorBase, unittest.TestCase):

    @property
    def reader(self):
        from ms_deisotope.data_source.thermo_raw import ThermoRawLoader
        return ThermoRawLoader(self.path)


@unittest.skipIf(missing_net_dll, "Requires .NET Runtime and RawFileReader.dll")
class TestNETThermoRawLoaderScanBehavior(ThermoRawLoaderScanBehaviorBase, unittest.TestCase):

    @property
    def reader(self):
        from ms_deisotope.data_source.thermo_raw_net import ThermoRawLoader
        return ThermoRawLoader(self.path)


if __name__ == '__main__':
    unittest.main()
